/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.control.task;

import java.util.ResourceBundle;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.FileOutputStream;
import java.io.PrintWriter;
import javax.swing.JOptionPane;
import javax.xml.transform.TransformerException;
import javax.xml.transform.Result;
import javax.xml.transform.Transformer;
import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.apache.xerces.dom.DocumentImpl;

import org.xmloperator.util.exception.ExceptionControler;
import org.xmloperator.framework.option.Constants;
import org.xmloperator.framework.option.ThreeStateOption;
import org.xmloperator.framework.option.Options;
import org.xmloperator.framework.document.transform.TransformerInfoXMLResultPart;
import org.xmloperator.framework.document.transform.TransformerInfo;
import org.xmloperator.framework.view.instance.ToolView;
import org.xmloperator.framework.view.instance.ViewAccess;
import org.xmloperator.framework.view.resource.Label;
import org.xmloperator.framework.control.task.CancelableTask;
import org.xmloperator.util.io.FileUtils;
import org.xmloperator.util.io.MIME2Java;
import org.xmloperator.util.view.ExceptionDialogs;
import org.xmloperator.document.util.DocumentWriter;
import org.xmloperator.control.io.IOUtils;

public final class DocumentTransformTask implements CancelableTask {
  private Transformer transformer;
  private TransformerInfo transformerInfo;
  private Source source;
  private String destinationPathName;
  private Options options;
  private ViewAccess viewAccess;

  private static final String getString(String key) {
    return ResourceBundle.getBundle(Label.CLASS).getString(key);
  }

  public DocumentTransformTask(
    Transformer transformer, TransformerInfo transformerInfo,
    Source source, String destinationPathName, Options options,
    ViewAccess viewAccess) {
    if (transformer == null || transformerInfo == null
      || source == null || destinationPathName == null
      || options == null || viewAccess == null)
      throw new IllegalArgumentException();
    this.transformer = transformer;
    this.transformerInfo = transformerInfo;
    this.source = source;
    this.destinationPathName = destinationPathName;
    this.options = options;
    this.viewAccess = viewAccess;
  }

  private boolean transformDocument() {
    displayMessage(getString(Label.TRANSFORMATION_IN_PROGRESS)
                   + " \"" + transformerInfo.getTransformerName() + "\"");
    try {
      if (transformerInfo.isResultXML()) {
        Document destinationDocument = new DocumentImpl();
        transformer.transform(source, new DOMResult(destinationDocument));
        displayMessage(getString(Label.ENREGISTREMENT_EN_COURS)
                       + " \"" + destinationPathName + "\"");
        TransformerInfoXMLResultPart xMLResultPart =
            transformerInfo.getXMLResultPart();
        String encoding = xMLResultPart.getEncodingName();
        if (encoding == null || encoding.length() == 0)
          encoding = options.getStringOptionValue(Options.DEFAULTENCODING);
        if (MIME2Java.convert(encoding) == null)
          encoding = Constants.DEFAULT_ENCODING;
        PrintWriter out = new PrintWriter(new OutputStreamWriter(
            new FileOutputStream(destinationPathName),
            MIME2Java.convert(encoding)));
        DocumentWriter documentWriter = new DocumentWriter(out);
        documentWriter.setShortcutUsedForEmptyElement(
            xMLResultPart.isShortcutUsedForEmptyElement());
        documentWriter.setLineFeedAtStartTagEnd(
            options.getThreeStateOptionValue(Options.ADDLINEFEEDATSTARTTAGEND)
            == ThreeStateOption.TRUE);
        documentWriter.printDocument(destinationDocument, null,
            xMLResultPart.getResultDTDPublicId(),
            xMLResultPart.getResultDTDSystemId(),
            null, encoding, xMLResultPart.isStandaloneDocument(), false);
        out.close();
      }
      else {
        PrintWriter out = new PrintWriter(new OutputStreamWriter(
            new FileOutputStream(destinationPathName)));
        transformer.transform(source, new StreamResult(out));
        out.close();
      }
      String systemId = FileUtils.reducedSystemId(destinationPathName);
      IOUtils.addResourceFirst(options, systemId);
      clearMessage();
      return true;
    }
    catch (TransformerException exception) {
      ExceptionDialogs.displayException("Transformer exception", exception);
    }
    catch (IOException exception) {
      ExceptionDialogs.displayException(exception);
    }
    clearMessage();
    return false;
  }

  private void displayMessage(String message) {
    ToolView toolView = viewAccess.getToolView();
    if (toolView != null)
      toolView.getToolStatusBar().displayMessage(message);
  }

  private void clearMessage() {
    displayMessage(" ");
  }

  ////////////////////////////////
  // CancelableTask implementation
  ////////////////////////////////

  public boolean run() {
    return transformDocument();
  }

  public void cancel() {
  }
}
