/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.control.task;

import java.util.ResourceBundle;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.LinkedList;
import java.util.HashSet;

import org.xmloperator.framework.view.resource.Label;
import org.xmloperator.framework.control.task.Task;
import org.xmloperator.framework.control.task.CancelableTask;
import org.xmloperator.framework.control.task.TaskControl;
import org.xmloperator.framework.control.global.GlobalControl;
import org.xmloperator.util.exception.ExceptionControler;


public final class TaskRunnerThread extends Thread implements TaskControl {
  private GlobalControl globalControl = null;
  private final LinkedList taskList = new LinkedList();
  private final HashSet refreshingTaskSet = new HashSet();
  private final Task disablingTask =
     new Task() {
       public boolean run() {
         globalControl.setEnabled(false);
         return true;
       }
     };
  private final Task enablingTask =
     new Task() {
       public boolean run() {
         globalControl.setEnabled(true);
         return true;
       }
     };
  private static final int STATE_WAITING = 0;
  private static final int STATE_DISABLING = 1;
  private static final int STATE_RUNNING_DISABLED = 2;
  private static final int STATE_ENABLING = 3;
  private static final int STATE_RUNNING_ENABLED = 4;
  private static final int STATE_REFRESHING = 5;
  private int state = STATE_WAITING;
  private Task submitedTask = null;
  private CancelableTask cancelableTask = null;
  private boolean isExceptionReported = false;

  private static final String getString(String key) {
    return ResourceBundle.getBundle(Label.CLASS).getString(key);
  }

  public TaskRunnerThread() {
    super("TaskRunner");
    this.setPriority(Thread.NORM_PRIORITY);
  }

  public void setGlobalControl(GlobalControl globalControl) {
    this.globalControl = globalControl;
  }

  //////////////////////////
  // Runnable implemantation
  //////////////////////////

  public void run() {
    try {
      while (true) {
        Task task = getNextTak();
        yield();
        runTask(task);
      }
    }
    catch (Exception exception) {
      ExceptionControler.printStackTrace(exception);
      System.exit(1);
    }
  }

  /**
   * Return null after one second at most for refreshing.
   */
  private Task getNextTak() {
    Task task = null;
    synchronized(taskList) {
      switch (state) {
        case STATE_DISABLING:
          state = STATE_RUNNING_DISABLED;
        case STATE_RUNNING_DISABLED:
          try {
            cancelableTask = null;
            task = (Task)taskList.removeFirst();
            if (CancelableTask.class.isAssignableFrom(task.getClass()))
              cancelableTask = (CancelableTask)task;
          }
          catch (NoSuchElementException exception) {
            state = STATE_ENABLING;
            task = enablingTask;
          }
          break;
        case STATE_ENABLING:
        case STATE_RUNNING_ENABLED:
        case STATE_REFRESHING:
          state = STATE_WAITING;
        case STATE_WAITING:
          if (submitedTask == null && taskList.isEmpty())
            try {
              taskList.wait(1000);
            }
            catch (InterruptedException exception) {
            }
          if (submitedTask != null) {
            state = STATE_RUNNING_ENABLED;
            task = submitedTask;
            submitedTask = null;
          }
          else if (!taskList.isEmpty()) {
            state = STATE_DISABLING;
            task = disablingTask;
          }
          else {
            state = STATE_REFRESHING;
            checkException();
          }
          break;
      }
    }
    return task;
  }

  private void checkException() {
    if (ExceptionControler.hasExceptionOccured())
      globalControl.displayMessage(getString(Label.SEE_EXCEPTION));
  }

  /**
   * task == null for refreshing.
   */
  private void runTask(Task task) {
    boolean canProcessingContinue = true;
    try {
      if (task != null)
        canProcessingContinue = task.run();
      else {
        Iterator iterator = refreshingTaskSet.iterator();
        while (iterator.hasNext()) {
          canProcessingContinue = ((Task)iterator.next()).run();
          if (!canProcessingContinue)
            break;
        }
      }
    }
    catch (Exception exception) {
      ExceptionControler.printStackTrace(exception);
      canProcessingContinue = false;
    }
    if (!canProcessingContinue)
      synchronized(taskList) {
        taskList.clear();
      }
  }

  /////////////////////////////
  // TaskControl implementation
  /////////////////////////////

  public void appendTask(Task task) {
    if (task != null)
      synchronized(taskList) {
        taskList.addLast(task);
        taskList.notify();
      }
  }

  public void cancelTasks() {
    synchronized(taskList) {
      taskList.clear();
    }
    CancelableTask cancelableTask = this.cancelableTask;
    if (cancelableTask != null)
      try {
        cancelableTask.cancel();
      }
      catch (Exception exception) {
        ExceptionControler.printStackTrace(exception);
      }
  }

  public boolean submitTask(Task task) {
    synchronized(taskList) {
      if (state != STATE_WAITING)
        return false;
      submitedTask = task;
      taskList.notify();
      return true;
    }
  }

  public void addRefreshingTask(Task task) {
    synchronized(refreshingTaskSet) {
      refreshingTaskSet.add(task);
    }
  }

  public void removeRefreshingTask(Task task) {
    synchronized(refreshingTaskSet) {
      refreshingTaskSet.remove(task);
    }
  }
}
