/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.document.manage;

import org.w3c.dom.Node;
import org.w3c.dom.Attr;
import org.w3c.dom.Element;

import org.xmloperator.framework.document.namespace.UndefinedNamespaceException;
import org.xmloperator.framework.document.id.IdException;
import org.xmloperator.framework.document.id.IdManager;
import org.xmloperator.framework.document.schema.ElementContent;
import org.xmloperator.framework.document.schema.AttributesContent;
import org.xmloperator.framework.document.schema.
  NotConformantToContentModelException;
import org.xmloperator.framework.document.manage.ElementDefinitionMapper;
import org.xmloperator.framework.document.manage.ElementOperationListener;
import org.xmloperator.framework.document.manage.AttributeChangeListener;
import org.xmloperator.framework.document.manage.AttributesContentAccess;
import org.xmloperator.document.util.FragmentModifier;
import org.xmloperator.document.manage.ContentCache;

public abstract class AbstractElementManager implements ElementDefinitionMapper,
  ElementOperationListener, AttributeChangeListener {
  protected IdManager idManager;
  protected final ContentCache contentCache = new ContentCache();

  public AbstractElementManager(IdManager idManager) {
    if (idManager == null)
      throw new IllegalArgumentException();
    this.idManager = idManager;
  }

  protected abstract void addElementContentToCache(Element parentElement)
    throws UndefinedNamespaceException, NotConformantToContentModelException;

  protected abstract void addAttributesContentToCache(Element parentElement);

  protected void removeElementDefinitionFromCache(Element element) {
  }

  private final void removeElementFromCache(Element element,
    boolean isDefinitionIncluded) {
    Node childNode = element.getFirstChild();
    while (childNode != null) {
      if (childNode.getNodeType() == Node.ELEMENT_NODE)
        this.removeElementFromCache((Element)childNode, true);
      childNode = childNode.getNextSibling();
    }
    this.contentCache.remove(element);
    if (isDefinitionIncluded)
      this.removeElementDefinitionFromCache(element);
  }

  private final void removeParentElementFromCache(Element childElement,
    boolean isDefinitionIncluded) {
    Node parentNode = childElement.getParentNode();
    if (parentNode != null && parentNode.getNodeType() == Node.ELEMENT_NODE)
      this.removeElementFromCache((Element)parentNode, isDefinitionIncluded);
  }

  /////////////////////////////////////////
  // AttributesContentAccess implementation
  /////////////////////////////////////////

  public final AttributesContent getAttributesContent(Element element) {
    this.addAttributesContentToCache(element);
    return contentCache.getAttributesContent(element);
  }

  /////////////////////////////////////////////////
  // ElementDefinitionMapper partial implementation
  /////////////////////////////////////////////////

  public final ElementContent getElementContent(Element parentElement)
    throws NotConformantToContentModelException {
    try {
      this.addElementContentToCache(parentElement);
    }
    catch (UndefinedNamespaceException exception) {
      return null;
    }
    return contentCache.getElementContent(parentElement);
  }

  public final boolean isElementContentValid(Element parentElement) {
    try {
      return getElementContent(parentElement) != null;
    }
    catch (NotConformantToContentModelException exception) {
      return false;
    }
  }

  public final void removeElementFromCache(Element element) {
    removeElementFromCache(element, true);
  }

  //////////////////////////////////////////
  // ElementOperationListener implementation
  //////////////////////////////////////////

  public final void notifyElementInsert(Element element) {
    if (element == null)
      throw new IllegalArgumentException();
    this.removeParentElementFromCache(element, false);
    if (!idManager.isEmpty())
      this.changeIdWhenNecessary(element);
    this.noteIds(element);
    this.noteRefs(element);
  }

  public final void notifyElementsInsert(Element[] elements) {
    for (int k = 0; k < elements.length; k++) {
      Element element = elements[k];
      if (element == null)
        throw new IllegalArgumentException();
      this.removeParentElementFromCache(element, false);
    }
    for (int k = 0; k < elements.length; k++)
      this.changeIdWhenNecessary(elements[k]);
    for (int k = 0; k < elements.length; k++)
      this.noteIds(elements[k]);
    for (int k = 0; k < elements.length; k++)
      this.noteRefs(elements[k]);
  }

  public final void notifyElementRemove(Element element) {
    if (element == null)
      throw new IllegalArgumentException();
    this.freeRefs(element);
    this.freeIds(element);
    this.removeParentElementFromCache(element, false);
  }

  public final void notifyElementsRemove(Element[] elements) {
    for (int k = 0; k < elements.length; k++)
      this.freeRefs(elements[k]);
    for (int k = 0; k < elements.length; k++)
      this.freeIds(elements[k]);
    for (int k = 0; k < elements.length; k++) {
      Element element = elements[k];
      if (element == null)
        throw new IllegalArgumentException();
      this.removeParentElementFromCache(element, false);
    }
  }

  public final void notifyBeforeElementMove(Element element) {
    this.removeParentElementFromCache(element, false);
  }

  public final void notifyBeforeElementsMove(Element[] elements) {
    for (int k = 0; k < elements.length; k++)
      this.removeParentElementFromCache(elements[k], false);
  }

  public final void notifyAfterElementMove(Element element) {
    this.removeParentElementFromCache(element, false);
  }

  public final void notifyAfterElementsMove(Element[] elements) {
    for (int k = 0; k < elements.length; k++)
      this.removeParentElementFromCache(elements[k], false);
  }

  private void changeIdWhenNecessary(Element element) {
    FragmentModifier.changeIdWhenNecessary(
      (AttributesContentAccess)this, idManager, element);
  }

  private void noteIds(Element element) {
    {
      AttributesContent attributesContent = this.getAttributesContent(element);
      if (attributesContent != null) {
        String idAttrName = attributesContent.getIdAttributeName();
        if (idAttrName != null) {
          Attr idAttribute = element.getAttributeNode(idAttrName);
          if (idAttribute != null)
            try {
              idManager.addId(idAttribute, idAttribute.getValue());
            }
            catch (IdException exception) {
              idAttribute.setValue("");
            }
        }
      }
    }
    {
      Node childNode = element.getFirstChild();
      while (childNode != null) {
        if (childNode.getNodeType() == Node.ELEMENT_NODE)
          noteIds((Element)childNode);
        childNode = childNode.getNextSibling();
      }
    }
  }

  private void noteRefs(Element element) {
    {
      AttributesContent attributesContent = this.getAttributesContent(element);
      if (attributesContent != null)
        for (int k = 0; k < attributesContent.getPossibleNameCount(); k++) {
          String attrName = attributesContent.getPossibleName(k);
          String attrType = attributesContent.getType(attrName);
          if (attrType.indexOf(AttributesContent.TYPE_IDREF) == 0) {
            Attr idRefAttribute = element.getAttributeNode(attrName);
            if (idRefAttribute != null)
              try {
                if (attrType.equals(AttributesContent.TYPE_IDREF))
                  idManager.addIdRef(idRefAttribute);
                else if (attrType.equals(AttributesContent.TYPE_IDREFS))
                  idManager.addIdRefs(idRefAttribute);
              }
              catch (IdException exception) {
                idRefAttribute.setValue("");
              }
          }
        }
    }
    {
      Node childNode = element.getFirstChild();
      while (childNode != null) {
        if (childNode.getNodeType() == Node.ELEMENT_NODE)
          noteRefs((Element)childNode);
        childNode = childNode.getNextSibling();
      }
    }
  }

  private void freeIds(Element element) {
    {
      Node childNode = element.getLastChild();
      while (childNode != null) {
        if (childNode.getNodeType() == Node.ELEMENT_NODE)
          freeIds((Element)childNode);
        childNode = childNode.getPreviousSibling();
      }
    }
    {
      AttributesContent attributesContent = this.getAttributesContent(element);
      if (attributesContent != null) {
        String idAttrName = attributesContent.getIdAttributeName();
        if (idAttrName != null) {
          Attr idAttribute = element.getAttributeNode(idAttrName);
          if (idAttribute != null)
            idManager.removeId(idAttribute, idAttribute.getValue());
        }
      }
    }
  }

  private void freeRefs(Element element) {
    {
      Node childNode = element.getLastChild();
      while (childNode != null) {
        if (childNode.getNodeType() == Node.ELEMENT_NODE)
          freeRefs((Element)childNode);
        childNode = childNode.getPreviousSibling();
      }
    }
    {
      AttributesContent attributesContent = this.getAttributesContent(element);
      if (attributesContent != null)
        for (int k = 0; k < attributesContent.getPossibleNameCount(); k++) {
          String attrName = attributesContent.getPossibleName(k);
          String attrType = attributesContent.getType(attrName);
          if (attrType.indexOf(AttributesContent.TYPE_IDREF) == 0) {
            Attr idRefAttribute = element.getAttributeNode(attrName);
            if (idRefAttribute != null) {
              if (attrType.equals(AttributesContent.TYPE_IDREF))
                idManager.removeIdRef(idRefAttribute);
              else if (attrType.equals(AttributesContent.TYPE_IDREFS))
                idManager.removeIdRefs(idRefAttribute);
            }
          }
        }
    }
  }

  /////////////////////////////////////////
  // AttributeChangeListener implementation
  /////////////////////////////////////////

  public final void notifyAttributeChange(Element element,
    boolean isNamespaceDeclaration) {
    if (isNamespaceDeclaration)
      this.removeParentElementFromCache(element, false);
    else
      this.removeElementFromCache(element, false);
  }
}
