/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.document.manage;

import java.util.List;

import org.w3c.dom.Node;
import org.w3c.dom.Text;
import org.w3c.dom.Element;

import org.xmloperator.framework.document.manage.ElementOperationListener;
import org.xmloperator.framework.document.manage.TreeViewDocumentListener;
import org.xmloperator.framework.document.manage.NodeChangeEvent;
import org.xmloperator.framework.document.manage.NodeChangeListener;
import org.xmloperator.framework.document.manage.DocumentModifiedListener;
import org.xmloperator.framework.document.manage.NodeManager;
import org.xmloperator.util.exception.ExceptionControler;
import org.xmloperator.document.manage.NodeChangeEventImpl;

public final class NodeManagerImpl implements NodeManager {
  private final List nodeChangeListenerList;
  private final List documentModifiedListenerList;
  private TreeViewDocumentListener treeViewDocumentListener = null;
  private ElementOperationListener elementOperationListener = null;

  public NodeManagerImpl(
      final List nodeChangeListenerList,
      final List documentModifiedListenerList) {
    this.nodeChangeListenerList = nodeChangeListenerList;
    this.documentModifiedListenerList = documentModifiedListenerList;
  }

  public void setTreeViewDocumentListener(
      TreeViewDocumentListener treeViewDocumentListener) {
    this.treeViewDocumentListener = treeViewDocumentListener;
  }

  public void setElementOperationListener(
      ElementOperationListener elementOperationListener) {
    this.elementOperationListener = elementOperationListener;
  }

  private void notifyNodeChangeEvent(NodeChangeEvent nodeChangeEvent) {
    if (this.nodeChangeListenerList != null)
      for (int k = 0; k < this.nodeChangeListenerList.size(); k++)
        ((NodeChangeListener)this.nodeChangeListenerList.get(k)).
            notifyNodeChangeEvent(nodeChangeEvent);
  }

  private void notifyPossibleAppearanceChange(Node node) {
    if (node != null)
      this.notifyNodeChangeEvent(
          NodeChangeEventImpl.nodeAppearanceMayHaveChangedEvent(node));
  }

  private void notifyPossibleAppearanceChangeIncludingParent(Node node) {
    if (node != null) {
      this.notifyPossibleAppearanceChange(node);
      this.notifyPossibleAppearanceChange(node.getParentNode());
    }
  }

  private void notifyDocumentModified() {
    if (this.documentModifiedListenerList != null)
      for (int k = 0; k < this.documentModifiedListenerList.size(); k++)
        ((DocumentModifiedListener)this.documentModifiedListenerList.get(k)).
            notifyDocumentModified();
  }

  private void processExceptionPostAction(Exception exception) {
    ExceptionControler.printStackTrace(exception);
  }

  /////////////////////////////
  // NodeManager implementation
  /////////////////////////////

  public void insertNode(Node parentNode, Node nextSibling, Node newNode) {
    parentNode.insertBefore(newNode, nextSibling);
    try {
      if (this.elementOperationListener != null
          && newNode.getNodeType() == Node.ELEMENT_NODE)
        this.elementOperationListener.notifyElementInsert( (Element) newNode);
      this.notifyDocumentModified();
      this.notifyPossibleAppearanceChangeIncludingParent(parentNode);
      this.treeViewDocumentListener.notifyNodeInsert(newNode);
    }
    catch (Exception exception) {
      this.processExceptionPostAction(exception);
    }
  }

  public void insertNodes(
      Node parentNode, Node nextSibling, Node[] newNodes) {
    for (int k = 0; k < newNodes.length; k++)
      parentNode.insertBefore(newNodes[k], nextSibling);
    try {
      if (this.elementOperationListener != null) {
        Element[] elements = elements(newNodes);
        if (elements != null)
          this.elementOperationListener.notifyElementsInsert(elements);
      }
      this.notifyDocumentModified();
      this.notifyPossibleAppearanceChangeIncludingParent(parentNode);
      this.treeViewDocumentListener.notifyNodesInsert(newNodes);
    }
    catch (Exception exception) {
      this.processExceptionPostAction(exception);
    }
  }

  public void removeNode(Node node) {
    Node parentNode = node.getParentNode();
    Node nodeToSelect = node.getNextSibling();
    if (nodeToSelect == null) {
      nodeToSelect = node.getPreviousSibling();
      if (nodeToSelect == null)
        nodeToSelect = parentNode;
    }
    Object context =
        this.treeViewDocumentListener.notifyBeforeNodeRemoveMoveOrReplace(node);
    this.notifyNodeChangeEvent(NodeChangeEventImpl.nodeRemoveEvent(node));
    if (this.elementOperationListener != null
      && node.getNodeType() == Node.ELEMENT_NODE)
      this.elementOperationListener.notifyElementRemove((Element)node);
    parentNode.removeChild(node);
    try {
      this.notifyDocumentModified();
      this.notifyPossibleAppearanceChangeIncludingParent(parentNode);
      this.treeViewDocumentListener.notifyAfterNodesRemove(
          context, nodeToSelect);
    }
    catch (Exception exception) {
      this.processExceptionPostAction(exception);
    }
  }

  public void removeNodes(Node[] nodes) {
    Node parentNode = nodes[0].getParentNode();
    Node nodeToSelect = nodes[nodes.length - 1].getNextSibling();
    if (nodeToSelect == null) {
      nodeToSelect = nodes[0].getPreviousSibling();
      if (nodeToSelect == null)
        nodeToSelect = parentNode;
    }
    Object context =
        this.treeViewDocumentListener.notifyBeforeNodesRemoveMoveOrReplace(nodes);
    for (int k = 0; k < nodes.length; k++)
      this.notifyNodeChangeEvent(NodeChangeEventImpl.nodeRemoveEvent(nodes[k]));
    if (elementOperationListener != null) {
      Element[] elements = elements(nodes);
      if (elements != null)
        elementOperationListener.notifyElementsRemove(elements);
    }
    for (int k = 0; k < nodes.length; k++)
      parentNode.removeChild(nodes[k]);
    try {
      this.notifyDocumentModified();
      this.notifyPossibleAppearanceChangeIncludingParent(parentNode);
      this.treeViewDocumentListener.notifyAfterNodesRemove(
          context, nodeToSelect);
    }
    catch (Exception exception) {
      this.processExceptionPostAction(exception);
    }
  }

  public void moveNode(
      Node node, Node destinationParentNode, Node destinationNextSibling) {
    Object context =
        this.treeViewDocumentListener.notifyBeforeNodeRemoveMoveOrReplace(node);
    Element element = this.elementOperationListener != null
        && node.getNodeType() == Node.ELEMENT_NODE ? (Element) node : null;
    if (element != null)
      this.elementOperationListener.notifyBeforeElementMove(element);
    Node sourceParentNode = node.getParentNode();
    destinationParentNode.insertBefore(node, destinationNextSibling);
    try {
      if (element != null)
        this.elementOperationListener.notifyAfterElementMove(element);
      this.notifyDocumentModified();
      this.notifyPossibleAppearanceChangeIncludingParent(sourceParentNode);
      this.notifyPossibleAppearanceChangeIncludingParent(destinationParentNode);
      this.treeViewDocumentListener.notifyAfterNodeMoveOrReplace(context, node);
    }
    catch (Exception exception) {
      this.processExceptionPostAction(exception);
    }
  }

  public void moveNodes(
      Node[] nodes, Node destinationParentNode, Node destinationNextSibling) {
    if (nodes.length == 0)
      return;
    Object context =
        this.treeViewDocumentListener.notifyBeforeNodesRemoveMoveOrReplace(
        nodes);
    Element[] elements = elementOperationListener != null ?
        elements(nodes) : null;
    if (elements != null)
      this.elementOperationListener.notifyBeforeElementsMove(elements);
    Node sourceParentNode = nodes[0].getParentNode();
    for (int k = 0; k < nodes.length; k++)
      destinationParentNode.insertBefore(nodes[k], destinationNextSibling);
    try {
      if (elements != null)
        this.elementOperationListener.notifyAfterElementsMove(elements);
      this.notifyDocumentModified();
      this.notifyPossibleAppearanceChangeIncludingParent(sourceParentNode);
      this.notifyPossibleAppearanceChangeIncludingParent(destinationParentNode);
      this.treeViewDocumentListener.notifyAfterNodesMoveOrReplace(
          context, nodes);
    }
    catch (Exception exception) {
      this.processExceptionPostAction(exception);
    }
  }

  public void replaceNodes(Node[] oldNodes, Node[] newNodes) {
    Node parentNode = oldNodes[0].getParentNode();
    Object context =
        this.treeViewDocumentListener.notifyBeforeNodesRemoveMoveOrReplace(
        oldNodes);
    for (int k = 0; k < oldNodes.length; k++)
      this.notifyNodeChangeEvent(NodeChangeEventImpl.nodeRemoveEvent(oldNodes[k]));
    if (elementOperationListener != null) {
      Element[] elements = elements(oldNodes);
      if (elements != null)
        elementOperationListener.notifyElementsRemove(elements);
    }
    if (oldNodes.length == 1 && newNodes.length == 1)
      parentNode.replaceChild(newNodes[0], oldNodes[0]);
    else {
      Node nextSibling = oldNodes[oldNodes.length - 1].getNextSibling();
      for (int k = 0; k < oldNodes.length; k++)
        parentNode.removeChild(oldNodes[k]);
      for (int k = 0; k < newNodes.length; k++)
        parentNode.insertBefore(newNodes[k], nextSibling);
    }
    try {
      if (this.elementOperationListener != null) {
        Element[] elements = elements(newNodes);
        if (elements != null)
          this.elementOperationListener.notifyElementsInsert(elements);
      }
      this.notifyDocumentModified();
      this.notifyPossibleAppearanceChangeIncludingParent(parentNode);
      this.treeViewDocumentListener.notifyAfterNodesMoveOrReplace(
          context, newNodes);
    }
    catch (Exception exception) {
      this.processExceptionPostAction(exception);
    }
  }

  public void setNodeValue(Node node, String newValue) {
    node.setNodeValue(newValue);
    try {
      this.notifyDocumentModified();
      this.notifyPossibleAppearanceChangeIncludingParent(node.getParentNode());
      this.treeViewDocumentListener.notifyNodeSetValue(node);
    }
    catch (Exception exception) {
      this.processExceptionPostAction(exception);
    }
  }

  public void joinNextText(Text textNodeToJoinToNextOne) {
    Node parentNode = textNodeToJoinToNextOne.getParentNode();
    Node nextSibling = textNodeToJoinToNextOne.getNextSibling();
    Object context =
        this.treeViewDocumentListener.notifyBeforeTextJoin(nextSibling);
    this.notifyNodeChangeEvent(NodeChangeEventImpl.nodeRemoveEvent(nextSibling));
    textNodeToJoinToNextOne.appendData(nextSibling.getNodeValue());
    parentNode.removeChild(nextSibling);
    try {
      this.notifyDocumentModified();
      this.notifyPossibleAppearanceChangeIncludingParent(parentNode);
      this.treeViewDocumentListener.notifyAfterTextJoin(
          context, textNodeToJoinToNextOne);
    }
    catch (Exception exception) {
      this.processExceptionPostAction(exception);
    }
  }

  public void splitText(
      Text textNodeToSplit, int offset, Text recipientForSecondPart) {
    Text newText = textNodeToSplit.splitText(offset);
    if (recipientForSecondPart != null) {
      recipientForSecondPart.setNodeValue(newText.getNodeValue());
      newText.getParentNode().replaceChild(recipientForSecondPart, newText);
      newText = recipientForSecondPart;
    }
    try {
      this.notifyDocumentModified();
      this.notifyPossibleAppearanceChangeIncludingParent(
          textNodeToSplit.getParentNode());
      this.treeViewDocumentListener.notifyTextSplit(textNodeToSplit);
    }
    catch (Exception exception) {
      this.processExceptionPostAction(exception);
    }
  }

  public void replaceElementName(Element oldElement, Element newElement) {
    if (newElement == oldElement)
      return;
    Object context =
        this.treeViewDocumentListener.notifyBeforeNodeRemoveMoveOrReplace(
        oldElement);
    this.notifyNodeChangeEvent(NodeChangeEventImpl.nodeRemoveEvent(oldElement));
    if (this.elementOperationListener != null)
      this.elementOperationListener.notifyElementRemove(oldElement);
    oldElement.getParentNode().replaceChild(newElement, oldElement);
    while (true) {
      Node childNode = oldElement.getFirstChild();
      if (childNode == null)
        break;
      newElement.appendChild(childNode);
    }
    try {
      if (this.elementOperationListener != null)
        this.elementOperationListener.notifyElementInsert(newElement);
      this.notifyDocumentModified();
      this.notifyPossibleAppearanceChange(newElement.getParentNode());
      this.treeViewDocumentListener.notifyAfterNodeMoveOrReplace(
          context, newElement);
    }
    catch (Exception exception) {
      this.processExceptionPostAction(exception);
    }
  }

  public void packNodes(Node[] nodes, Element newElement) {
    Object context =
        this.treeViewDocumentListener.notifyBeforeNodeIntervalPack(nodes);
    for (int k = 0; k < nodes.length; k++)
      this.notifyNodeChangeEvent(NodeChangeEventImpl.nodeRemoveEvent(nodes[k]));
    if (elementOperationListener != null) {
      Element[] elements = elements(nodes);
      if (elements != null)
        elementOperationListener.notifyElementsRemove(elements);
    }
    {
      Node firstNode = nodes[0];
      firstNode.getParentNode().insertBefore(newElement, firstNode);
    }
    for (int k = 0; k < nodes.length; k++)
      newElement.appendChild(nodes[k]);
    try {
      if (this.elementOperationListener != null)
        this.elementOperationListener.notifyElementInsert(newElement);
      this.notifyDocumentModified();
      this.notifyPossibleAppearanceChange(newElement.getParentNode());
      this.treeViewDocumentListener.notifyAfterNodeIntervalPack(
          context, newElement);
    }
    catch (Exception exception) {
      this.processExceptionPostAction(exception);
    }
  }

  public void unPackElement(Element element) {
    Object context =
        this.treeViewDocumentListener.notifyBeforeElementUnPack(element);
    this.notifyNodeChangeEvent(NodeChangeEventImpl.nodeRemoveEvent(element));
    if (this.elementOperationListener != null)
      this.elementOperationListener.notifyElementRemove(element);
    Node parentNode = element.getParentNode();
    Node[] children = new Node[childCount(element)];
    int index = 0;
    while (true) {
      Node childNode = element.getFirstChild();
      if (childNode == null)
        break;
      children[index++] = childNode;
      parentNode.insertBefore(childNode, element);
    }
    parentNode.removeChild(element);
    try {
      if (elementOperationListener != null) {
        Element[] elements = elements(children);
        if (elements != null)
          elementOperationListener.notifyElementsInsert(elements);
      }
      this.notifyDocumentModified();
      this.notifyPossibleAppearanceChange(parentNode);
      this.treeViewDocumentListener.notifyAfterElementUnpack(context, children);
    }
    catch (Exception exception) {
      this.processExceptionPostAction(exception);
    }
  }

  private static final int childCount(Element element) {
    int count = 0;
    Node childNode = element.getFirstChild();
    while (childNode != null) {
      count++;
      childNode = childNode.getNextSibling();
    }
    return count;
  }

  public void reorderNodes(Node[] nodesInDestinationOrder) {
    Node parentNode = nodesInDestinationOrder[0].getParentNode();
    Node destinationNextSibling =
        lastNode(nodesInDestinationOrder).getNextSibling();
    Object context =
        this.treeViewDocumentListener.notifyBeforeNodesRemoveMoveOrReplace(
        nodesInDestinationOrder);
    Element[] elements = elementOperationListener != null ?
        elements(nodesInDestinationOrder) : null;
    if (elements != null)
      this.elementOperationListener.notifyBeforeElementsMove(elements);
    for (int k = 0; k < nodesInDestinationOrder.length; k++)
      parentNode.insertBefore(
          nodesInDestinationOrder[k], destinationNextSibling);
    try {
      if (elements != null)
        this.elementOperationListener.notifyAfterElementsMove(elements);
      this.notifyDocumentModified();
      this.notifyPossibleAppearanceChange(parentNode);
      this.treeViewDocumentListener.notifyAfterNodesMoveOrReplace(
          context, nodesInDestinationOrder);
    }
    catch (Exception exception) {
      this.processExceptionPostAction(exception);
    }
  }

  private static final Node lastNode(Node[] nodes) {
    Node parentNode = nodes[0].getParentNode();
    int lastIndex = -1;
    Node lastNode = null;
    for (int k = 0; k < nodes.length; k++) {
      Node node = nodes[k];
      if (node.getParentNode() != parentNode)
        throw new IllegalArgumentException();
      int index = getNodeIndex(node);
      if (index > lastIndex) {
        lastIndex = index;
        lastNode = node;
      }
    }
    return lastNode;
  }

  private static final int getNodeIndex(Node node) {
    int index = 0;
    while (true) {
      node = node.getPreviousSibling();
      if (node == null)
        return index;
      index++;
    }
  }

  private static final Element[] elements(Node[] nodes) {
    int count = 0;
    for (int k = 0; k < nodes.length; k++)
      if (nodes[k].getNodeType() == Node.ELEMENT_NODE)
        count++;
    if (count == 0)
      return null;
    Element[] elements = new Element[count];
    count = 0;
    for (int k = 0; k < nodes.length; k++) {
      Node node = nodes[k];
      if (node.getNodeType() == Node.ELEMENT_NODE)
        elements[count++] = (Element)node;
    }
    return elements;
  }
}
