/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.document.relaxng;

import java.util.Random;
import java.util.Arrays;
import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;

import org.w3c.dom.Node;
import org.w3c.dom.Element;

import org.xmloperator.framework.document.namespace.QualifiedName;
import org.xmloperator.framework.document.relaxng.NameClass;
import org.xmloperator.framework.document.relaxng.Name;
import org.xmloperator.framework.document.relaxng.ChoiceNameClass;
import org.xmloperator.framework.document.relaxng.Name;
import org.xmloperator.framework.document.relaxng.NsName;
import org.xmloperator.framework.document.relaxng.Pattern;
import org.xmloperator.framework.document.relaxng.NonEmptyPattern;
import org.xmloperator.framework.document.relaxng.Data;
import org.xmloperator.framework.document.relaxng.Value;
import org.xmloperator.framework.document.relaxng.GroupPattern;
import org.xmloperator.framework.document.relaxng.Attribute;
import org.xmloperator.framework.document.schema.AttributesContent;
import org.xmloperator.util.dom.QualifiedNameImpl;
import org.xmloperator.util.dom.QualifiedNameUtils;
import org.xmloperator.util.dom.NamespaceUtils;
import org.xmloperator.document.relaxng.NameClassUtils;

// Help for selecting a valid set of attribute names for an element.
public final class AttributesContentImpl implements AttributesContent {
  private Element element;
  private NonEmptyPattern rootPattern;
  private QualifiedName[] qualifiedNames;
  private String[] attrNames;
  private final Map fromAttrNameToType = new HashMap();
  private final Map fromAttrNameToEnumeration = new HashMap();
  private final Set requiredAttrNames = new HashSet();
  private String idAttrName = null;
  private static final int tryCountByNameSetDistance = 16;
  private Random random = new Random(0);

  // Constructs an AttributesContent for an element given by its root pattern.
  public AttributesContentImpl(Element element, NonEmptyPattern rootPattern) {
    if (element == null || rootPattern == null)
      throw new IllegalArgumentException();
    this.element = element;
    this.rootPattern = rootPattern;
    Set qualifiedNameSet = new HashSet();
    appendPatternToAttributeQualifiedNameSet(qualifiedNameSet, rootPattern);
    qualifiedNames = new QualifiedNameImpl[qualifiedNameSet.size()];
    qualifiedNameSet.toArray(qualifiedNames);
    Arrays.sort(qualifiedNames);
    attrNames = new String[qualifiedNames.length];
    for (int k = 0; k < attrNames.length; k++) {
      QualifiedName qualifiedName = qualifiedNames[k];
      String attrName =
          QualifiedNameUtils.getPseudoAttributeName(qualifiedName, element);
      attrNames[k] = attrName;
      Attribute attribute = getNamedAttribute(rootPattern,
        qualifiedName.getNamespaceURI(), qualifiedName.getLocalName());
      if (attribute == null)
        continue;
      if (isNamedAttributeRequired(rootPattern, attribute))
        requiredAttrNames.add(attrName);
      Pattern attributePattern = attribute.getPattern();
      if (attributePattern.isEmpty())
        continue;
      switch (((NonEmptyPattern)attributePattern).getPatternType()) {
        case NonEmptyPattern.PATTERN_DATA: {
          String idType = idType((Data)attributePattern);
          if (idType != null) {
            fromAttrNameToType.put(attrName, idType);
            if (idType.equals(TYPE_ID))
              idAttrName = attrName;
          }
          break;
        }
        case NonEmptyPattern.PATTERN_GROUP: {
          String[] enumeration = enumeration((GroupPattern)attributePattern);
          if (enumeration != null) {
            fromAttrNameToType.put(attrName, TYPE_ENUMERATION);
            fromAttrNameToEnumeration.put(attrName, enumeration);
          }
          break;
        }
        case NonEmptyPattern.PATTERN_VALUE: {
          fromAttrNameToType.put(attrName, TYPE_ENUMERATION);
          fromAttrNameToEnumeration.put(attrName,
            new String[] {((Value)attributePattern).getValue()});
          break;
        }
      }
    }
  }

  private static final void appendPatternToAttributeQualifiedNameSet(
    Set attributeQualifiedNameSet, NonEmptyPattern pattern) {
    switch (pattern.getPatternType()) {
      case NonEmptyPattern.PATTERN_ATTRIBUTE:
        NameClassUtils.appendNameClassToQualifiedNameSet(
          attributeQualifiedNameSet, ((Attribute)pattern).getNameClass());
        break;
      case NonEmptyPattern.PATTERN_GROUP:
        for (int k = 0; k < ((GroupPattern)pattern).getChildCount(); k++)
          appendPatternToAttributeQualifiedNameSet(attributeQualifiedNameSet,
            ((GroupPattern)pattern).getChild(k));
        break;
    }
  }

  private static final Attribute getNamedAttribute(NonEmptyPattern pattern,
    String namespaceURI, String localName) {
    switch (pattern.getPatternType()) {
      case NonEmptyPattern.PATTERN_ATTRIBUTE: {
        NameClass nameClass = ((Attribute)pattern).getNameClass();
        return nameClass instanceof Name
          && ((Name)nameClass).getName().equals(localName)
          && ((Name)nameClass).getNsURI().equals(namespaceURI) ?
          (Attribute)pattern : null;
      }
      case NonEmptyPattern.PATTERN_GROUP:
        for (int k = 0; k < ((GroupPattern)pattern).getChildCount(); k++) {
          Attribute attribute = getNamedAttribute(
            ((GroupPattern)pattern).getChild(k), namespaceURI, localName);
          if (attribute != null)
            return attribute;
        }
        return null;
      default:
        return null;
    }
  }

  private static final boolean isNamedAttributeRequired(NonEmptyPattern pattern,
    Attribute namedAttribute) {
    switch (pattern.getPatternType()) {
      case NonEmptyPattern.PATTERN_ATTRIBUTE:
        return pattern == namedAttribute;
      case NonEmptyPattern.PATTERN_GROUP:
        int groupType = ((GroupPattern)pattern).getGroupType();
        if (groupType == GroupPattern.GROUP_CHOICE_INCLUDING_EMPTY_PATTERN)
          return false;
        int childCount = ((GroupPattern)pattern).getChildCount();
        if (groupType == GroupPattern.GROUP_CHOICE_EXCLUDING_EMPTY_PATTERN
          && childCount > 1)
          return false;
        for (int k = 0; k < childCount; k++) {
          if (isNamedAttributeRequired(((GroupPattern)pattern).getChild(k),
            namedAttribute))
            return true;
        }
        return false;
      default:
        return false;
    }
  }

  private static final String idType(Data data) {
    if (!data.getDatatypeLibraryURI().
      equals("http://relaxng.org/ns/compatibility/datatypes/1.0"))
      return null;
    String dataType = data.getType();
    if (dataType.equals("ID"))
      return TYPE_ID;
    if (dataType.equals("IDREF"))
      return TYPE_IDREF;
    if (dataType.equals("IDREFS"))
      return TYPE_IDREFS;
    return null;
  }

  private static final String[] enumeration(GroupPattern enumerationPattern) {
    switch (enumerationPattern.getGroupType()) {
      case GroupPattern.GROUP_CHOICE_EXCLUDING_EMPTY_PATTERN:
      case GroupPattern.GROUP_CHOICE_INCLUDING_EMPTY_PATTERN:
        break;
      default:
        return null;
    }
    List valueList = new ArrayList();
    for (int k = 0; k < enumerationPattern.getChildCount(); k++)
      {
      NonEmptyPattern childPattern = enumerationPattern.getChild(k);
      if (childPattern.getPatternType() != NonEmptyPattern.PATTERN_VALUE)
        return null;
      valueList.add(((Value)childPattern).getValue());
    }
    String[] values = new String[valueList.size()];
    valueList.toArray(values);
    return values;
  }

  ///////////////////////////////////
  // AttributesContent implementation
  ///////////////////////////////////

  public int getPossibleNameCount() {
    return attrNames.length;
  }

  public String getPossibleName(int index) {
    return attrNames[index];
  }

  public String getType(String attrName) {
    String type = (String)fromAttrNameToType.get(attrName);
    return type != null ? type : TYPE_CDATA;
  }

  public String[] getEnumeration(String attrName) {
    return (String[])fromAttrNameToEnumeration.get(attrName);
  }

  public boolean isRequired(String attrName) {
    return requiredAttrNames.contains(attrName);
  }

  public String getIdAttributeName() {
    return idAttrName;
  }

  //////////////////////////////
  // Ideas for future extensions
  //////////////////////////////

  // Return the nearest valid set of attributes names
  //   from a given set of attribute names.
  // Return the wishedNameSet itself if valid.
  // This may fail if the root pattern is ambiguous or not deterministic.
  // Name sets are relative to the list of all possible names.
  // They have (must have for wishedNameSet) the possible name count as length.
  // The distance between a valid name set and the wished name set
  //   is define as the count of alone names.
  // An alone name is a name which belongs to one set but not to the other set.
  // An exact result (if exists) is not guaranted if the wishedNameSet
  //   is not a valid name set.
  // Throws a NullPointerException if any name is accepted.
  // UNUSED FOR THE MOMENT. JUST AN IDEA.
  ///////////////////////////////////////
  public boolean[] nearestValidNameSet(boolean[] wishedNameSet) {
    if (wishedNameSet == null || wishedNameSet.length != qualifiedNames.length)
      throw new IllegalArgumentException();
    boolean[] nameSet = new boolean[wishedNameSet.length];
    int distanceMax = 0;
    while (true) {
      for (int k = 0; k < tryCountByNameSetDistance; k++) {
        Arrays.fill(nameSet, false);
        makeValidNameSet(nameSet, wishedNameSet, rootPattern);
        int distance = distanceBetweenNameSets(nameSet, wishedNameSet);
        if (distance <= distanceMax)
          return distance == 0 ? wishedNameSet : nameSet;
      }
      distanceMax++;
    }
  }

  private void makeValidNameSet(boolean[] nameSet,
    boolean[] wishedNameSet, NonEmptyPattern pattern) {
    switch (pattern.getPatternType()) {
      case NonEmptyPattern.PATTERN_ATTRIBUTE:
        makeValidNameSet(nameSet, wishedNameSet,
          ((Attribute)pattern).getNameClass());
        break;
      case NonEmptyPattern.PATTERN_GROUP:
        int childCount = ((GroupPattern)pattern).getChildCount();
        int groupType = ((GroupPattern)pattern).getGroupType();
        switch (groupType) {
          case GroupPattern.GROUP_CHOICE_INCLUDING_EMPTY_PATTERN:
          case GroupPattern.GROUP_CHOICE_EXCLUDING_EMPTY_PATTERN:
            boolean isRequired =
              groupType == GroupPattern.GROUP_CHOICE_EXCLUDING_EMPTY_PATTERN;
            for (int k = 0; k < childCount; k++) {
              NonEmptyPattern childPattern =
                ((GroupPattern)pattern).getChild(k);
              int favor = getPatternFavor(wishedNameSet, childPattern);
              if (favor > 0) {
                makeValidNameSet(nameSet, wishedNameSet, childPattern);
                return;
              }
              isRequired &= favor < 0;
            }
            if (isRequired)
              makeValidNameSet(nameSet, wishedNameSet,
                ((GroupPattern)pattern).getChild(random.nextInt(childCount)));
            break;
          default:
            for (int k = 0; k < childCount; k++)
              makeValidNameSet(nameSet, wishedNameSet,
                ((GroupPattern)pattern).getChild(k));
        }
        break;
    }
  }

  private void makeValidNameSet(boolean[] nameSet, boolean[] wishedNameSet,
    NameClass nameClass) {
    switch (nameClass.getNameClassType()) {
      case NameClass.NAMECLASS_CHOICE:
        int childCount = ((ChoiceNameClass)nameClass).getChildCount();
        for (int k = 0; k < childCount; k++) {
          NameClass childNameClass = ((ChoiceNameClass)nameClass).getChild(k);
          if (isNameClassFavorable(wishedNameSet, childNameClass)) {
            makeValidNameSet(nameSet, wishedNameSet, childNameClass);
            return;
          }
        }
        makeValidNameSet(nameSet, wishedNameSet,
          ((ChoiceNameClass)nameClass).getChild(random.nextInt(childCount)));
        return;
      case NameClass.NAMECLASS_NAME:
        nameSet[indexOfQualifiedName(qualifiedNames, new QualifiedNameImpl(
          ((Name)nameClass).getNsURI(), ((Name)nameClass).getName()))] = true;
        return;
      case NameClass.NAMECLASS_NSNAME:
        nameSet[indexOfQualifiedName(qualifiedNames, new QualifiedNameImpl(
          ((NsName)nameClass).getNsURI(), ""))] = true;
        return;
    }
  }

  // Return +1 if the pattern is favorable, -1 if unfavorable,
  //   0 if the pattern doesn't contain a required attribute pattern.
  private int getPatternFavor(boolean[] wishedNameSet,
    NonEmptyPattern pattern) {
    switch (pattern.getPatternType()) {
      case NonEmptyPattern.PATTERN_ATTRIBUTE:
        return isNameClassFavorable(wishedNameSet,
          ((Attribute)pattern).getNameClass()) ? +1 : -1;
      case NonEmptyPattern.PATTERN_GROUP:
        int childCount = ((GroupPattern)pattern).getChildCount();
        int groupType = ((GroupPattern)pattern).getGroupType();
        switch (groupType) {
          case GroupPattern.GROUP_CHOICE_INCLUDING_EMPTY_PATTERN:
          case GroupPattern.GROUP_CHOICE_EXCLUDING_EMPTY_PATTERN: {
            boolean isUnfavorable =
              groupType == GroupPattern.GROUP_CHOICE_EXCLUDING_EMPTY_PATTERN;
            for (int k = 0; k < childCount; k++) {
              int favor = getPatternFavor(wishedNameSet,
                ((GroupPattern)pattern).getChild(k));
              if (favor > 0)
                return +1;
              isUnfavorable &= favor < 0;
            }
            return isUnfavorable ? -1 : 0;
          }
          default: {
            boolean isFavorable = false;
            for (int k = 0; k < childCount; k++) {
              int favor = getPatternFavor(wishedNameSet,
                ((GroupPattern)pattern).getChild(k));
              if (favor < 0)
                return -1;
              isFavorable |= favor > 0;
            }
            return isFavorable ? +1 : 0;
          }
        }
      default:
        return 0;
    }
  }

  private boolean isNameClassFavorable(boolean[] wishedNameSet,
    NameClass nameClass) {
    switch (nameClass.getNameClassType()) {
      case NameClass.NAMECLASS_NAME:
        return wishedNameSet[indexOfQualifiedName(qualifiedNames,
          new QualifiedNameImpl(((Name)nameClass).getNsURI(),
          ((Name)nameClass).getName()))];
      case NameClass.NAMECLASS_NSNAME:
        return wishedNameSet[indexOfQualifiedName(qualifiedNames,
          new QualifiedNameImpl(((NsName)nameClass).getNsURI(), ""))];
      case NameClass.NAMECLASS_CHOICE:
        for (int k = 0; k < ((ChoiceNameClass)nameClass).getChildCount(); k++)
          if (isNameClassFavorable(wishedNameSet,
            ((ChoiceNameClass)nameClass).getChild(k)))
            return true;
        return false;
      default:
        return false;
    }
  }

  private static final int indexOfQualifiedName(QualifiedName[] qualifiedNames,
    QualifiedName qualifiedName) {
    int index = Arrays.binarySearch(qualifiedNames, qualifiedName);
    if (index < 0 || !qualifiedNames[index].equals(qualifiedName))
      throw new RuntimeException("Pattern changed during attributes selecting");
    return index;
  }

  private int distanceBetweenNameSets(boolean[] nameSet1, boolean[] nameSet2) {
    int length = nameSet1.length;
    if (nameSet2.length != length)
      throw new IllegalArgumentException();
    int count = 0;
    for (int k = 0; k < length; k++)
      if (nameSet1[k] != nameSet2[k])
        count++;
    return count;
  }

  // Adds and/or removes attributes to element in such a way that its
  //   attribute names sets conforms to nameSet.
  // Doesn't add or remove NsName attributes.
  // Return false if fails to find prefix mapping to namespace.
  // Throws a NullPointerException if any name is accepted.
  // UNUSED BECAUSE NOT A REVERSIBLE OPERATION. JUST FOR DESIGN.
  //////////////////////////////////////////////////////////////
  public boolean conformElementToAttributeNameSet(Element element,
    boolean[] nameSet) {
    if (nameSet.length != qualifiedNames.length)
      throw new IllegalArgumentException();
    boolean isConformant = true;
    for (int k = 0; k < qualifiedNames.length; k++) {
      QualifiedName qualifiedName = qualifiedNames[k];
      String localName = qualifiedName.getLocalName();
      if (localName.length() != 0) {
        String namespaceURI = qualifiedName.getNamespaceURI();
        boolean isAttributeExisting =
          element.getAttributeNodeNS(namespaceURI, localName) != null;
        if (nameSet[k] != isAttributeExisting)
          if (isAttributeExisting)
            element.removeAttributeNS(namespaceURI, localName);
          else if (namespaceURI.length() == 0)
            element.setAttributeNS("", localName, "");
          else {
            String prefix =
              NamespaceUtils.namespaceToPrefix(element, namespaceURI);
            if (prefix != null)
              element.setAttributeNS(namespaceURI, prefix + ":" + localName,
                "");
            else
              isConformant = false;
          }
      }
    }
    return isConformant;
  }
}
