/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.document.relaxng;

import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.HashSet;

import org.w3c.dom.Element;

import org.xmloperator.framework.document.relaxng.NameClass;
import org.xmloperator.framework.document.relaxng.Name;
import org.xmloperator.framework.document.relaxng.Pattern;
import org.xmloperator.framework.document.relaxng.NonEmptyPattern;
import org.xmloperator.framework.document.relaxng.Attribute;
import org.xmloperator.framework.document.relaxng.GroupPattern;
import org.xmloperator.framework.document.relaxng.ElementDefinition;
import org.xmloperator.framework.document.relaxng.Ref;
import org.xmloperator.framework.document.namespace.QualifiedName;
import org.xmloperator.framework.document.schema.ElementContent;
import org.xmloperator.framework.document.schema.
    NotConformantToContentModelException;
import org.xmloperator.framework.document.schema.UnsatisfiedPatternException;
import org.xmloperator.framework.document.schema.
    InvalidElementSpecificationException;
import org.xmloperator.framework.document.schema.InvalidAttributesException;
import org.xmloperator.util.dom.QualifiedNameImpl;
import org.xmloperator.document.relaxng.PatternUtils;

// May be not RELAX NG compliant if the content model is ambiguous
//   or not deterministic. For instance, the content (a, b) is rejected
//   if the content model is (a?, a, b).
public final class ElementContentImpl implements ElementContent {

  // Return true if the content is valid against the content model.
  // The returned value is undefined if the content model is ambiguous
  //   or not deterministic.
  // Each contentElementSpec must be either an QualifiedName or an
  //   ElementDefinition.
  public static final boolean isContentValid(
    final NonEmptyPattern rootPattern, final Object[] contentElementSpecs,
    final QualifiedName[] attributeSpecs) {
    try {
      new ElementContentImpl(rootPattern, contentElementSpecs, attributeSpecs);
      return true;
    }
    catch (NotConformantToContentModelException exception) {
      return false;
    }
  }

  private NonEmptyPattern rootPattern;
  private int deepth;
  private Object[] contentElementSpecs;
  private int[][] patternPaths;
  private ElementDefinition[] elementDefinitions;

  public ElementContentImpl(final NonEmptyPattern rootPattern,
    final Object[] contentElementSpecs, final QualifiedName[] attributeSpecs)
    throws NotConformantToContentModelException {
    if (rootPattern == null || contentElementSpecs == null
        || attributeSpecs == null)
      throw new IllegalArgumentException();
    for (int k = 0; k < contentElementSpecs.length; k++) {
      Object contentElementSpec = contentElementSpecs[k];
      if (contentElementSpec == null
          || (!(contentElementSpec instanceof QualifiedName)
          && !(contentElementSpec instanceof ElementDefinition)))
        throw new IllegalArgumentException();
    }
    this.rootPattern = rootPattern;
    this.deepth = PatternUtils.computeDeepth(rootPattern);
    this.contentElementSpecs = contentElementSpecs;
    int contentElementSpecCount = contentElementSpecs.length;
    this.patternPaths = new int[contentElementSpecCount][deepth + 1];
    this.elementDefinitions = new ElementDefinition[contentElementSpecCount];
    this.computePatternPaths(attributeSpecs);
  }

  private void computePatternPaths(QualifiedName[] attributeSpecs)
    throws NotConformantToContentModelException {
    attributeSpecs = (QualifiedName[])attributeSpecs.clone();
    final int[][] patternUseStack = new int[deepth][];
    Arrays.fill(patternUseStack, null);
    final PatternPathComputer patternPathComputer =
      new PatternPathComputer(attributeSpecs, patternUseStack);
    final int[] attributePatternPath = new int[deepth + 1];
    for (int k = 0; k < contentElementSpecs.length; k++)
      try {
        Object contentElementSpec = contentElementSpecs[k];
        int[] patternPath = patternPaths[k];
        Arrays.fill(patternPath, -1);
        patternPathComputer.setElementSpec(contentElementSpec, patternPath);
        while (!patternPathComputer.computePatternPath(rootPattern)) {
          System.arraycopy(patternPath, 0, attributePatternPath, 0, deepth + 1);
          patternPathComputer.setPreviousPatternPath(attributePatternPath);
        }
        elementDefinitions[k] =
          ((Ref)PatternUtils.getPatternFromPath(rootPattern, patternPath)).
          getElementDefinition();
        patternPathComputer.setPreviousPatternPath(patternPath);
      }
      catch (NotConformantToContentModelException exception) {
        exception.contentElementSpecIndex = k;
        throw exception;
      }
    try { // is there an unsatisfied pattern ?
      int[] patternPath = new int[deepth + 1];
      Arrays.fill(patternPath, -1);
      patternPathComputer.setElementSpec(QualifiedNameImpl.EMPTY_NAME,
          patternPath);
      while (!patternPathComputer.computePatternPath(rootPattern)) {
        System.arraycopy(patternPath, 0, attributePatternPath, 0, deepth + 1);
        patternPathComputer.setPreviousPatternPath(attributePatternPath);
      }
    }
    catch (InvalidElementSpecificationException exception) {
      // Normal. EMPTY_NAME is not valid. There is no unsatisfied pattern.
    }
    if (notNullCount(attributeSpecs) != 0)
      throw new InvalidAttributesException(
          notNullQualifiedNames(attributeSpecs));
  }

  private static final int notNullCount(Object[] objects) {
    int count = 0;
    for (int k = 0; k < objects.length; k++)
      if (objects[k] != null)
        count++;
    return count;
  }

  private static final QualifiedName[] notNullQualifiedNames(
      QualifiedName[] qualifiedNames) {
    int notNullCount = notNullCount(qualifiedNames);
    QualifiedName[] notNullQualifiedNames = new QualifiedName[notNullCount];
    int index = 0;
    for (int k = 0; k < qualifiedNames.length; k++) {
      QualifiedName qualifiedName = qualifiedNames[k];
      if (qualifiedName != null)
        notNullQualifiedNames[index++] = qualifiedName;
    }
    return notNullQualifiedNames;
  }

  private static class PatternPathComputer {
    private static final int PATTERN_IS_NOT_USED = 0;
    private static final int PATTERN_IS_USED = 1;
    private static final int PATTERN_IS_CONSUMED = 2;

    private static final int REPORT__PATTERN_IGNORABLE = 0;
    private static final int REPORT__ATTRIBUTE_FOUND = 1;
    private static final int REPORT__ELEMENT_DEFINITION_FOUND = 2;
    private static final int
      REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_CONSUMED = 3;
    private static final int
      REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_UNCONSUMED = 4;
    private static final int REPORT__PATTERN_UNSATISFIED = 5;

    private static final InvalidElementSpecificationException
      INVALID_ELEMENT_SPECIFICATION_EXCEPTION =
      new InvalidElementSpecificationException();

    private QualifiedName[] attributeSpecs;
    private Object contentElementSpec = null;
    private boolean isElementSpecAnElementDefinition = false;
    private int[] patternPath = null;
    private int[] previousPatternPath = null;
    private int[][] patternUseStack;
    private final List unsatisfiedPatternList = new ArrayList();

    public PatternPathComputer(QualifiedName[] attributeSpecs,
      int[][] patternUseStack) {
      this.attributeSpecs = attributeSpecs;
      this.patternUseStack = patternUseStack;
    }

    public void setElementSpec(Object contentElementSpec, int[] patternPath) {
      this.contentElementSpec = contentElementSpec;
      this.isElementSpecAnElementDefinition =
        contentElementSpec instanceof ElementDefinition;
      this.patternPath = patternPath;
    }

    public void setPreviousPatternPath(int[] previousPatternPath) {
      this.previousPatternPath = previousPatternPath;
    }

    private UnsatisfiedPatternException createUnsatisfiedPatternException() {
      NonEmptyPattern[] unsatisfiedPatterns =
        new NonEmptyPattern[unsatisfiedPatternList.size()];
      unsatisfiedPatternList.toArray(unsatisfiedPatterns);
      return new UnsatisfiedPatternException(unsatisfiedPatterns);
    }

    // Computes patternPath from rootPattern for succedding
    //   to previousPatternPath if any
    //   or raises a NotConformantToContentModelException.
    // Return false if an attribute has been found.
    public boolean computePatternPath(final NonEmptyPattern rootPattern)
      throws NotConformantToContentModelException {
      unsatisfiedPatternList.clear();
      switch (previousPatternPath != null ?
        computePatternPathSucceeding(rootPattern, 0, false) :
        computePatternPath(rootPattern, 0)) {
          case REPORT__PATTERN_UNSATISFIED:
            throw createUnsatisfiedPatternException();
          case REPORT__PATTERN_IGNORABLE:
            throw INVALID_ELEMENT_SPECIFICATION_EXCEPTION;
          case REPORT__ATTRIBUTE_FOUND:
            return false;
          default:
            return true;
        }
    }

    // Compute the pattern path that is the first
    //   to succeed to the previous one and to map to the contentElementSpec.
    // Raises an UnsatisfiedPatternException if a pattern is unsatisfied.
    // Return REPORT__PATTERN_IGNORABLE
    //   or REPORT__ATTRIBUTE_FOUND
    //   or REPORT__ELEMENT_DEFINITION_FOUND
    private int computePatternPathSucceeding(
      final NonEmptyPattern pattern, final int startingIndex,
      boolean canRepeatPreviousPath) throws UnsatisfiedPatternException {
      if (!canRepeatPreviousPath)
        canRepeatPreviousPath = pattern.canBeRepeated();
      switch (pattern.getPatternType()) {
        case NonEmptyPattern.PATTERN_ATTRIBUTE: {
          if (!canRepeatPreviousPath)
            return REPORT__PATTERN_IGNORABLE;
          if (consumeOneAttribute(((Attribute)pattern).getNameClass()))
            return REPORT__ATTRIBUTE_FOUND;
          return REPORT__PATTERN_IGNORABLE;
        }
        case NonEmptyPattern.PATTERN_REF: {
          if (!canRepeatPreviousPath)
            return REPORT__PATTERN_IGNORABLE;
          ElementDefinition elementDefinition =
              ((Ref)pattern).getElementDefinition();
          if (isElementSpecAnElementDefinition ?
              elementDefinition != contentElementSpec :
              !isNameContained(elementDefinition.getNameClass(),
              (QualifiedName)contentElementSpec)
              || elementDefinition.getPattern() == null)
            return REPORT__PATTERN_IGNORABLE;
          patternPath[startingIndex] = -1;
          return REPORT__ELEMENT_DEFINITION_FOUND;
        }
        case NonEmptyPattern.PATTERN_GROUP: {
          int groupType = ((GroupPattern)pattern).getGroupType();
          int childCount = ((GroupPattern)pattern).getChildCount();
          int previousPathStep = previousPatternPath[startingIndex];
          {
            patternPath[startingIndex] = previousPathStep;
            boolean isChoice =
                groupType == GroupPattern.GROUP_CHOICE_INCLUDING_EMPTY_PATTERN
                || groupType
                == GroupPattern.GROUP_CHOICE_EXCLUDING_EMPTY_PATTERN;
            switch (computePatternPathSucceeding(
                ((GroupPattern)pattern).getChild(previousPathStep),
                startingIndex + 1, isChoice & canRepeatPreviousPath)) {
              case REPORT__ATTRIBUTE_FOUND:
                return REPORT__ATTRIBUTE_FOUND;
              case REPORT__ELEMENT_DEFINITION_FOUND:
                return REPORT__ELEMENT_DEFINITION_FOUND;
            }
          }
          switch (groupType) {
            case GroupPattern.GROUP_GROUP: {
              for (int k = previousPathStep + 1; k < childCount; k++) {
                patternPath[startingIndex] = k;
                NonEmptyPattern childPattern = ((GroupPattern)pattern).getChild(k);
                switch (computePatternPath(childPattern, startingIndex + 1)) {
                  case REPORT__ATTRIBUTE_FOUND:
                    return REPORT__ATTRIBUTE_FOUND;
                  case REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_CONSUMED:
                  case REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_UNCONSUMED:
                    return REPORT__ELEMENT_DEFINITION_FOUND;
                  case REPORT__PATTERN_UNSATISFIED:
                    throw createUnsatisfiedPatternException();
                }
              }
              break;
            }
            case GroupPattern.GROUP_INTERLEAVE: {
              int[] patternUse = patternUseStack[startingIndex];
              if (patternUse == null)
                throw new NullPointerException();
              boolean isThereAnUnsatisfiedPattern = false;
              for (int k = 0; k < childCount; k++)
                if (k != previousPathStep
                  && patternUse[k] != PATTERN_IS_CONSUMED) {
                patternPath[startingIndex] = k;
                NonEmptyPattern childPattern =
                    ((GroupPattern)pattern).getChild(k);
                int unsatisfiedPatternCount = unsatisfiedPatternList.size();
                switch (computePatternPath(childPattern,  startingIndex + 1)) {
                  case REPORT__ATTRIBUTE_FOUND:
                    patternUse[k] = PATTERN_IS_CONSUMED;
                    return REPORT__ATTRIBUTE_FOUND;
                  case REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_CONSUMED:
                    patternUse[k] = PATTERN_IS_CONSUMED;
                    return REPORT__ELEMENT_DEFINITION_FOUND;
                  case REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_UNCONSUMED:
                    patternUse[k] = PATTERN_IS_USED;
                    return REPORT__ELEMENT_DEFINITION_FOUND;
                  case REPORT__PATTERN_UNSATISFIED:
                    if (patternUse[k] == PATTERN_IS_NOT_USED)
                      isThereAnUnsatisfiedPattern = true;
                    else
                      truncateList(unsatisfiedPatternList,
                                   unsatisfiedPatternCount);
                    break;
                }
              }
              if (isThereAnUnsatisfiedPattern)
                throw createUnsatisfiedPatternException();
              break;
            }
          }
          if (canRepeatPreviousPath) {
            int unsatisfiedPatternCount = unsatisfiedPatternList.size();
            switch (computePatternPath(pattern, startingIndex)) {
              case REPORT__ATTRIBUTE_FOUND:
                return REPORT__ATTRIBUTE_FOUND;
              case REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_CONSUMED:
              case REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_UNCONSUMED:
                return REPORT__ELEMENT_DEFINITION_FOUND;
              case REPORT__PATTERN_UNSATISFIED:
                truncateList(unsatisfiedPatternList, unsatisfiedPatternCount);
                break;
            }
          }
          return REPORT__PATTERN_IGNORABLE;
        }
        default:
          return REPORT__PATTERN_IGNORABLE;
      }
    }

    // Compute the pattern path that is the first to map to the qualified name.
    // Return REPORT__PATTERN_IGNORABLE
    //   or REPORT__ATTRIBUTE_FOUND
    //   or REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_CONSUMED
    //   or REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_UNCONSUMED
    //   or REPORT__PATTERN_UNSATISFIED
    private int computePatternPath(final NonEmptyPattern pattern,
      final int startingIndex) {
      switch (pattern.getPatternType()) {
        case NonEmptyPattern.PATTERN_ATTRIBUTE: {
          if (consumeOneAttribute(((Attribute)pattern).getNameClass()))
            return REPORT__ATTRIBUTE_FOUND;
          unsatisfiedPatternList.add(pattern);
          return REPORT__PATTERN_UNSATISFIED;
        }
        case NonEmptyPattern.PATTERN_REF: {
          ElementDefinition elementDefinition =
              ((Ref)pattern).getElementDefinition();
          if (isElementSpecAnElementDefinition ?
              elementDefinition != contentElementSpec :
              !isNameContained(elementDefinition.getNameClass(),
              (QualifiedName)contentElementSpec)
              || elementDefinition.getPattern() == null) {
            unsatisfiedPatternList.add(pattern);
            return REPORT__PATTERN_UNSATISFIED;
          }
          patternPath[startingIndex] = -1;
          return pattern.canBeRepeated() ?
              REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_UNCONSUMED :
              REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_CONSUMED;
        }
        case NonEmptyPattern.PATTERN_GROUP: {
          int groupType = ((GroupPattern)pattern).getGroupType();
          int childCount = ((GroupPattern)pattern).getChildCount();
          int unsatisfiedPatternCount = unsatisfiedPatternList.size();
          int report = REPORT__PATTERN_IGNORABLE;
          switch (groupType) {
            case GroupPattern.GROUP_GROUP: {
              for (int k = 0; k < childCount; k++) {
                patternPath[startingIndex] = k;
                switch (computePatternPath(((GroupPattern)pattern).getChild(k),
                    startingIndex + 1)) {
                  case REPORT__ATTRIBUTE_FOUND:
                    return REPORT__ATTRIBUTE_FOUND;
                  case REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_CONSUMED:
                  case REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_UNCONSUMED:
                    return pattern.canBeRepeated() ?
                        REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_UNCONSUMED :
                        REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_CONSUMED;
                  case REPORT__PATTERN_UNSATISFIED:
                    return REPORT__PATTERN_UNSATISFIED;
                }
              }
              return REPORT__PATTERN_IGNORABLE;
            }
            case GroupPattern.GROUP_INTERLEAVE: {
              int[] patternUse = patternUseStack[startingIndex];
              if (patternUse == null || patternUse.length < childCount) {
                patternUse = new int[childCount];
                patternUseStack[startingIndex] = patternUse;
              }
              Arrays.fill(patternUse, 0, childCount, PATTERN_IS_NOT_USED);
              for (int k = 0; k < childCount; k++)
                if (patternUse[k] != PATTERN_IS_CONSUMED) {
                  patternPath[startingIndex] = k;
                  switch (computePatternPath(
                      ((GroupPattern)pattern).getChild(k), startingIndex + 1)) {
                    case REPORT__ATTRIBUTE_FOUND:
                      patternUse[k] = PATTERN_IS_CONSUMED;
                      return REPORT__ATTRIBUTE_FOUND;
                    case REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_CONSUMED:
                      patternUse[k] = PATTERN_IS_CONSUMED;
                      return pattern.canBeRepeated() ?
                          REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_UNCONSUMED :
                          REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_CONSUMED;
                    case REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_UNCONSUMED:
                      patternUse[k] = PATTERN_IS_USED;
                      return pattern.canBeRepeated() ?
                          REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_UNCONSUMED :
                          REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_CONSUMED;
                    case REPORT__PATTERN_UNSATISFIED:
                      report = REPORT__PATTERN_UNSATISFIED;
                  }
                }
                /*
                if (report == REPORT__PATTERN_UNSATISFIED)
                  report = REPORT__PATTERN_UNSATISFIED;
                */
                break;
            }
            default: {
              report = groupType ==
                  GroupPattern.GROUP_CHOICE_EXCLUDING_EMPTY_PATTERN ?
                  REPORT__PATTERN_UNSATISFIED : REPORT__PATTERN_IGNORABLE;
              for (int k = 0; k < childCount; k++) {
                patternPath[startingIndex] = k;
                switch (computePatternPath(
                    ((GroupPattern)pattern).getChild(k), startingIndex + 1)) {
                  case REPORT__ATTRIBUTE_FOUND:
                    truncateList(unsatisfiedPatternList,
                                 unsatisfiedPatternCount);
                    return REPORT__ATTRIBUTE_FOUND;
                  case REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_CONSUMED:
                  case REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_UNCONSUMED:
                    truncateList(unsatisfiedPatternList,
                                 unsatisfiedPatternCount);
                    return pattern.canBeRepeated() ?
                        REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_UNCONSUMED :
                        REPORT__ELEMENT_DEFINITION_FOUND__PATTERN_CONSUMED;
                  case REPORT__PATTERN_IGNORABLE:
                    report = REPORT__PATTERN_IGNORABLE;
                    break;
                }
              }
              /*
              if (report == REPORT__PATTERN_UNSATISFIED)
                report = REPORT__PATTERN_UNSATISFIED;
              */
            }
          }
          if (report != REPORT__PATTERN_UNSATISFIED)
            truncateList(unsatisfiedPatternList, unsatisfiedPatternCount);
          /*
          else
            report = REPORT__PATTERN_UNSATISFIED;
          */
          return report;
        }
        default:
          return REPORT__PATTERN_IGNORABLE;
      }
    }

    private static final void truncateList(List list, int count) {
      while (true) {
        int size = list.size();
        if (size <= count)
          return;
        list.remove(size - 1);
      }
    }

    private boolean consumeOneAttribute(NameClass nameClass) {
      for (int k = 0; k < attributeSpecs.length; k++) {
        QualifiedName qualifiedName = attributeSpecs[k];
        if (qualifiedName != null
          && isNameContained(nameClass, qualifiedName)) {
          attributeSpecs[k] = null;
          return true;
        }
      }
      return false;
    }
  }

  private static final boolean isNameContained(NameClass nameClass,
    QualifiedName qualifiedName) {
    return NameClassUtils.isNameContained(nameClass,
      qualifiedName.getNamespaceURI(), qualifiedName.getLocalName());
  }

  ////////////////////////////////
  // ElementContent implementation
  ////////////////////////////////

  public int getElementSpecCount() {
    return contentElementSpecs.length;
  }

  // Return an QualifiedName or an ElementDefinition.
  public Object getElementSpec(final int contentElementSpecIndex) {
    return contentElementSpecs[contentElementSpecIndex];
  }

  // Return an ElementDefinition.
  public Object getElementDefinition(final int contentElementSpecIndex) {
    return elementDefinitions[contentElementSpecIndex];
  }

  public boolean isElementDefinitionRepeatable(
      final int contentElementSpecIndex) {
    return PatternUtils.isPatternRepeatable(rootPattern,
      patternPaths[contentElementSpecIndex]);
  }

  public void addElementDefinitionsForReplace(
    final Set elementDefinitionSet, final int contentElementSpecIndex) {
    PatternUtils.addElementDefinitionsForReplace(elementDefinitionSet,
      rootPattern, patternPaths[contentElementSpecIndex]);
  }

  public Set elementDefinitionsInsertableBefore(
      final int contentElementSpecIndex) {
    Set elementDefinitionSet = new HashSet();
    boolean isElementBefore = contentElementSpecIndex > 0;
    boolean isElementAfter =
        contentElementSpecIndex < contentElementSpecs.length;
    int[] patternPath1 =
      isElementBefore ? patternPaths[contentElementSpecIndex - 1] : null;
    int[] patternPath2 =
      isElementAfter ? patternPaths[contentElementSpecIndex] : null;
    if (isElementBefore && PatternUtils.isPatternRepeatable(rootPattern,
        patternPath1)) {
      elementDefinitionSet.add(elementDefinitions[contentElementSpecIndex - 1]);
      PatternUtils.addElementDefinitionsForReplaceRepeatedUse(
          elementDefinitionSet, rootPattern, patternPath1);
    }
    if (isElementAfter && PatternUtils.isPatternRepeatable(rootPattern,
        patternPath2)) {
      elementDefinitionSet.add(elementDefinitions[contentElementSpecIndex]);
      PatternUtils.addElementDefinitionsForReplaceRepeatedUse(
          elementDefinitionSet, rootPattern, patternPath2);
    }
    PatternUtils.addElementDefinitionsInsertableBetween(elementDefinitionSet,
      rootPattern, patternPath1, patternPath2);
    return elementDefinitionSet;
  }

  public String elementFromContentModelToString(
      final int contentElementSpecIndex, final boolean isSummarized,
      final Element contextElement) {
    return PatternUtils.childrenPatternToString(rootPattern,
      patternPaths[contentElementSpecIndex], isSummarized, contextElement);
  }
}
