/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.document.relaxng;

import java.util.Set;

import org.w3c.dom.Element;

import org.xmloperator.framework.document.relaxng.Pattern;
import org.xmloperator.framework.document.relaxng.NonEmptyPattern;
import org.xmloperator.framework.document.relaxng.Attribute;
import org.xmloperator.framework.document.relaxng.Ref;
import org.xmloperator.framework.document.relaxng.GroupPattern;
import org.xmloperator.framework.document.relaxng.ElementDefinition;
import org.xmloperator.document.relaxng.NameClassUtils;

public abstract class PatternUtils {

  //////////////////////////////////////////////////////////////////
  // Methods for geting general information about a children pattern
  //////////////////////////////////////////////////////////////////

  // return a string representing the children pattern
  public static final String childrenPatternToString(
      final Pattern pattern, final Element contextElement) {
    if (pattern.isEmpty())
      return "EMPTY";
    final StringBuffer s = new StringBuffer();
    appendChildrenPatternToString(s, (NonEmptyPattern)pattern, contextElement);
    merge_OneOrMore_and_Optional(s);
    merge_ZeroOrMore_and_Optional(s);
    return s.toString();
  }

  private static final void merge_OneOrMore_and_Optional(StringBuffer s) {
    int index = 0;
    while (index + 1 < s.length()) {
      if (s.charAt(index) == '+' && s.charAt(index + 1) == '?')
        s.replace(index, index + 2, "*");
      index++;
    }
  }

  private static final void merge_ZeroOrMore_and_Optional(StringBuffer s) {
    int index = 0;
    while (index + 1 < s.length()) {
      if (s.charAt(index) == '*' && s.charAt(index + 1) == '?')
        s.deleteCharAt(index + 1);
      else
        index++;
    }
  }

  private static final void appendChildrenPatternToString(final StringBuffer s,
    final NonEmptyPattern pattern, final Element contextElement) {
    boolean isOptional = false;
    switch (pattern.getPatternType()) {
      case NonEmptyPattern.PATTERN_TEXT:
        s.append("TEXT");
        break;
      case NonEmptyPattern.PATTERN_DATA:
        s.append("DATA");
        break;
      case NonEmptyPattern.PATTERN_VALUE:
        s.append("VALUE");
        break;
      case NonEmptyPattern.PATTERN_LIST:
        s.append("LIST");
        break;
      case NonEmptyPattern.PATTERN_ATTRIBUTE:
        s.append('@');
        s.append(NameClassUtils.nameClassToString(
          ((Attribute)pattern).getNameClass(), contextElement));
        break;
      case NonEmptyPattern.PATTERN_REF:
        s.append(NameClassUtils.nameClassToString(
          ((Ref)pattern).getElementDefinition().getNameClass(),
          contextElement));
        break;
      case NonEmptyPattern.PATTERN_GROUP: {
        int childCount = ((GroupPattern)pattern).getChildCount();
        if (childCount > 1)
          s.append('(');
        int groupType = ((GroupPattern)pattern).getGroupType();
        isOptional =
          groupType == GroupPattern.GROUP_CHOICE_INCLUDING_EMPTY_PATTERN;
        char groupTypeChar = groupPatternTypeChar(groupType);
        for (int k = 0; k < childCount; k++) {
          if (k != 0) {
            s.append(' ');
            s.append(groupTypeChar);
            s.append(' ');
          }
          appendChildrenPatternToString(s, ((GroupPattern)pattern).getChild(k),
                                        contextElement);
        }
        if (childCount > 1)
          s.append(')');
        break;
      }
    }
    {
      boolean canBeRepeated = pattern.canBeRepeated();
      if (isOptional | canBeRepeated)
        s.append(isOptional ? canBeRepeated ? '*' : '?' : '+');
    }
  }

  private static final char groupPatternTypeChar(int groupPatternType) {
    switch (groupPatternType) {
      case GroupPattern.GROUP_GROUP:
        return ',';
      case GroupPattern.GROUP_INTERLEAVE:
        return '&';
      default:
        return '|';
    }
  }

  // return the maximun length of element name path in a non empty pattern
  public static final int computeDeepth(final NonEmptyPattern pattern) {
    if (pattern.getPatternType() != NonEmptyPattern.PATTERN_GROUP)
      return 0;
    int maxDeepth = 0;
    for (int k = 0; k < ((GroupPattern)pattern).getChildCount(); k++) {
      int deepth = 1 + computeDeepth(((GroupPattern)pattern).getChild(k));
      if (deepth > maxDeepth)
        maxDeepth = deepth;
    }
    return maxDeepth;
  }

  /////////////////
  // Boolean result
  /////////////////

  // Return true if the non empty pattern contains no element definition
  //   reference.
  public static final boolean hasOnlyAttributes(final NonEmptyPattern pattern) {
    switch (pattern.getPatternType()) {
      case NonEmptyPattern.PATTERN_ATTRIBUTE:
        return true;
      case NonEmptyPattern.PATTERN_GROUP:
        for (int k = 0; k < ((GroupPattern)pattern).getChildCount(); k++)
          if (!hasOnlyAttributes(((GroupPattern)pattern).getChild(k)))
            return false;
        return true;
      default:
        return false;
    }
  }

  // Return true if the pattern contains a text or text like pattern.
  public static final boolean canContainText(final NonEmptyPattern pattern) {
    switch (((NonEmptyPattern)pattern).getPatternType()) {
      case NonEmptyPattern.PATTERN_DATA:
      case NonEmptyPattern.PATTERN_TEXT:
      case NonEmptyPattern.PATTERN_VALUE:
      case NonEmptyPattern.PATTERN_LIST:
        return true;
      case NonEmptyPattern.PATTERN_GROUP:
        for (int k = 0; k < ((GroupPattern)pattern).getChildCount(); k++)
          if (canContainText(((GroupPattern)pattern).getChild(k)))
            return true;
      default:
        return false;
    }
  }

  // Return true if the pattern contains a required text or text like pattern
  //   and no element ref pattern.
  public static final boolean shouldContainText(final NonEmptyPattern pattern) {
    return hasRequiredText(pattern) && !hasElementRef(pattern);
  }

  private static final boolean hasRequiredText(final NonEmptyPattern pattern) {
    switch (((NonEmptyPattern)pattern).getPatternType()) {
      case NonEmptyPattern.PATTERN_DATA:
      case NonEmptyPattern.PATTERN_TEXT:
      case NonEmptyPattern.PATTERN_VALUE:
      case NonEmptyPattern.PATTERN_LIST:
        return true;
      case NonEmptyPattern.PATTERN_GROUP:
        switch (((GroupPattern)pattern).getGroupType()) {
          case GroupPattern.GROUP_GROUP:
          case GroupPattern.GROUP_INTERLEAVE:
            for (int k = 0; k < ((GroupPattern)pattern).getChildCount(); k++)
              if (hasRequiredText(((GroupPattern)pattern).getChild(k)))
                return true;
            return false;
          default:
            for (int k = 0; k < ((GroupPattern)pattern).getChildCount(); k++)
              if (!hasRequiredText(((GroupPattern)pattern).getChild(k)))
                return false;
            return true;
        }
      default:
        return false;
    }
  }

  private static final boolean hasElementRef(final NonEmptyPattern pattern) {
    switch (((NonEmptyPattern)pattern).getPatternType()) {
      case NonEmptyPattern.PATTERN_REF:
        return ((Ref)pattern).getElementDefinition().getPattern() != null;
      case NonEmptyPattern.PATTERN_GROUP:
        for (int k = 0; k < ((GroupPattern)pattern).getChildCount(); k++)
          if (hasElementRef(((GroupPattern)pattern).getChild(k)))
            return true;
      default:
        return false;
    }
  }

  ////////////////////////////////////////////////////////////////////////
  // Methods about geting the required or not required element definitions
  //   or attributes referenced by a pattern
  ////////////////////////////////////////////////////////////////////////

  // Adds to a set all the allowed element definitions referenced
  //   by a non empty pattern.
  public static final void addElementDefinitions(
    final Set elementDefinitionSet, final NonEmptyPattern pattern) {
    addElementDefinitions(elementDefinitionSet, pattern, false);
  }

  // Adds to a set all the required element definitions referenced
  //   by a non empty pattern.
  // Returns true if one element definition reference or one attribute,
  //   at least is required.
  public static final boolean addRequiredElementDefinitions(
    final Set elementDefinitionSet, final NonEmptyPattern pattern) {
    return addElementDefinitions(elementDefinitionSet, pattern, true);
  }

  // Adds to a set all the allowed element definitions referenced
  //   by a non empty pattern.
  // If the boolean condition argument is set
  //   then only the required element definitions are concerned.
  // Returns true if one element definition reference or one attribute,
  //   at least is required.
  private static final boolean addElementDefinitions(
    final Set elementDefinitionSet, final NonEmptyPattern pattern,
    boolean areRequiredElementDefinitionsOnlyConcerned) {
    switch (pattern.getPatternType()) {
      case NonEmptyPattern.PATTERN_REF: {
        ElementDefinition elementDefinition =
          ((Ref)pattern).getElementDefinition();
        if (elementDefinition.getPattern() != null)
          elementDefinitionSet.add(elementDefinition);
        return true;
      }
      case NonEmptyPattern.PATTERN_ATTRIBUTE:
        return true;
      case NonEmptyPattern.PATTERN_GROUP: {
        boolean hasRequiredElementOrAttributeRef = true;
        switch (((GroupPattern)pattern).getGroupType()) {
          case GroupPattern.GROUP_CHOICE_INCLUDING_EMPTY_PATTERN:
            hasRequiredElementOrAttributeRef = false;
            if (areRequiredElementDefinitionsOnlyConcerned)
              break;
          case GroupPattern.GROUP_CHOICE_EXCLUDING_EMPTY_PATTERN:
            for (int k = 0; k < ((GroupPattern)pattern).getChildCount(); k++)
              hasRequiredElementOrAttributeRef &= addElementDefinitions(
                elementDefinitionSet, ((GroupPattern)pattern).getChild(k),
                areRequiredElementDefinitionsOnlyConcerned);
            break;
          default:
            hasRequiredElementOrAttributeRef = false;
            for (int k = 0; k < ((GroupPattern)pattern).getChildCount(); k++)
              hasRequiredElementOrAttributeRef |= addElementDefinitions(
                elementDefinitionSet, ((GroupPattern)pattern).getChild(k),
                areRequiredElementDefinitionsOnlyConcerned);
            break;
        }
        return hasRequiredElementOrAttributeRef;
      }
      default:
        return false;
    }
  }

  // Adds to a set all the allowed attributes referenced
  //   by a non empty pattern.
  // If the boolean condition argument is set
  //   then only the required attributes are concerned.
  // Returns true if one element definition reference or one attribute,
  //   at least is required.
  public static final boolean addAttributes(
    final Set attributeSet, final NonEmptyPattern pattern,
    boolean areRequiredAttributesOnlyConcerned) {
    switch (pattern.getPatternType()) {
      case NonEmptyPattern.PATTERN_REF:
        return true;
      case NonEmptyPattern.PATTERN_ATTRIBUTE:
        attributeSet.add(pattern);
        return true;
      case NonEmptyPattern.PATTERN_GROUP: {
        boolean hasRequiredElementOrAttributeRef = true;
        switch (((GroupPattern)pattern).getGroupType()) {
          case GroupPattern.GROUP_CHOICE_INCLUDING_EMPTY_PATTERN:
            hasRequiredElementOrAttributeRef = false;
            if (areRequiredAttributesOnlyConcerned)
              break;
          case GroupPattern.GROUP_CHOICE_EXCLUDING_EMPTY_PATTERN:
            for (int k = 0; k < ((GroupPattern)pattern).getChildCount(); k++)
              hasRequiredElementOrAttributeRef &= addAttributes(
                attributeSet, ((GroupPattern)pattern).getChild(k),
                areRequiredAttributesOnlyConcerned);
            break;
          case GroupPattern.GROUP_GROUP:
          case GroupPattern.GROUP_INTERLEAVE:
            hasRequiredElementOrAttributeRef = false;
            for (int k = 0; k < ((GroupPattern)pattern).getChildCount(); k++)
              hasRequiredElementOrAttributeRef |= addAttributes(
                attributeSet, ((GroupPattern)pattern).getChild(k),
                areRequiredAttributesOnlyConcerned);
            break;
        }
        return hasRequiredElementOrAttributeRef;
      }
      default:
        return false;
    }
  }

  //////////////////////////////
  // Methods about pattern paths
  //////////////////////////////
  // A pattern path is a sequence of pattern index values, terminated by -1.
  //////////////////////////////////////////////////////////////////////////

  // return a string representing a pattern focused on an element.
  public static final String childrenPatternToString(final Pattern pattern,
    final int[] patternPath, final boolean isSummarized,
    final Element contextElement) {
    if (pattern.isEmpty())
      return "EMPTY";
    final StringBuffer s = new StringBuffer();
    appendChildrenPatternToString(s, true, (NonEmptyPattern)pattern,
      patternPath, 0, isSummarized, contextElement);
    merge_OneOrMore_and_Optional(s);
    merge_ZeroOrMore_and_Optional(s);
    return s.toString();
  }

  private static final void appendChildrenPatternToString(final StringBuffer s,
    final boolean isRoot, final NonEmptyPattern pattern,
    final int[] patternPath, final int startIndex,
    final boolean isSummarized, final Element contextElement) {
    boolean isOptional = false;
    int patternType = pattern.getPatternType();
    if (patternType == NonEmptyPattern.PATTERN_GROUP) {
      final int childCount = ((GroupPattern)pattern).getChildCount();
      if (childCount > 1)
        s.append('(');
      final int groupType = ((GroupPattern)pattern).getGroupType();
      isOptional =
        groupType == GroupPattern.GROUP_CHOICE_INCLUDING_EMPTY_PATTERN;
      final String groupTypeString =
        " " + groupPatternTypeChar(groupType) + " ";
      final int patternPathStep =
        patternPath != null ? patternPath[startIndex] : -1;
      if (isSummarized && childCount > 2) {
        appendChildrenPatternToString(s, false,
          ((GroupPattern)pattern).getChild(0),
          patternPathStep == 0 ? patternPath : null, startIndex + 1,
          isSummarized, contextElement);
        final String summarizedGroup = "...";
        if (patternPathStep != 1) {
          s.append(groupTypeString);
          s.append(summarizedGroup);
        }
        if (patternPathStep >= 1) {
          s.append(groupTypeString);
          appendChildrenPatternToString(s, false,
            ((GroupPattern)pattern).getChild(patternPathStep), patternPath,
            startIndex + 1, isSummarized, contextElement);
          if (patternPathStep < childCount - 1) {
            s.append(groupTypeString);
            s.append(summarizedGroup);
          }
        }
      }
      else
        for (int k = 0; k < childCount; k++) {
          if (k != 0)
            s.append(groupTypeString);
          appendChildrenPatternToString(s, false,
            ((GroupPattern)pattern).getChild(k),
            patternPathStep == k ? patternPath : null, startIndex + 1,
            isSummarized, contextElement);
        }
      if (childCount > 1)
        s.append(')');
    }
    else {
      final String space = patternPath != null ? "    " : "";
      s.append(space);
      switch (patternType) {
        case NonEmptyPattern.PATTERN_TEXT:
          s.append("TEXT");
          break;
        case NonEmptyPattern.PATTERN_DATA:
          s.append("DATA");
          break;
        case NonEmptyPattern.PATTERN_VALUE:
          s.append("VALUE");
          break;
        case NonEmptyPattern.PATTERN_LIST:
          s.append("LIST");
          break;
        case NonEmptyPattern.PATTERN_ATTRIBUTE:
          s.append('@');
          s.append(NameClassUtils.nameClassToString(
            ((Attribute)pattern).getNameClass(), contextElement));
          break;
        case NonEmptyPattern.PATTERN_REF: {
          s.append(NameClassUtils.nameClassToString(
            ((Ref)pattern).getElementDefinition().getNameClass(),
            contextElement));
          break;
        }
      }
      s.append(space);
    }
    {
      boolean canBeRepeated = pattern.canBeRepeated();
      if (isOptional | canBeRepeated)
        s.append(isOptional ? canBeRepeated ? '*' : '?' : '+');
    }
  }

  // Return the pattern at the end of a pattern path
  public static final NonEmptyPattern getPatternFromPath(
      final NonEmptyPattern rootPattern, final int[] patternPath) {
    NonEmptyPattern pattern = rootPattern;
    int pathIndex = 0;
    while (true) {
      int childIndex = patternPath[pathIndex];
      if (childIndex < 0)
        return pattern;
      if (pattern.getPatternType() != NonEmptyPattern.PATTERN_GROUP)
        throw new IllegalArgumentException();
      pattern = ((GroupPattern)pattern).getChild(childIndex);
      pathIndex++;
    }
  }

  // Return true if the pattern referenced by a pattern path can be repeated.
  public static final boolean isPatternRepeatable(
      final NonEmptyPattern rootPattern, final int[] patternPath) {
    NonEmptyPattern pattern = rootPattern;
    boolean isRepeatable = false;
    int pathIndex = 0;
    while (true) {
      int childIndex = patternPath[pathIndex];
      boolean isPatternPathEnd = childIndex < 0;
      boolean isLeafPattern =
          pattern.getPatternType() != NonEmptyPattern.PATTERN_GROUP;
      if (isLeafPattern != isPatternPathEnd)
        // The pattern path end must map to a leaf pattern
        throw new IllegalArgumentException();
      if (!isRepeatable)
        isRepeatable = pattern.canBeRepeated();
      if (isPatternPathEnd)
        return isRepeatable;
      if (isRepeatable) {
        int groupType = ((GroupPattern)pattern).getGroupType();
        if (groupType != GroupPattern.GROUP_CHOICE_INCLUDING_EMPTY_PATTERN
            && groupType != GroupPattern.GROUP_CHOICE_EXCLUDING_EMPTY_PATTERN)
          for (int k = 0; k < ( (GroupPattern) pattern).getChildCount(); k++)
            if (k != childIndex
                && !isPatternOptional(((GroupPattern)pattern).getChild(k))) {
              isRepeatable = false;
              break;
            }
      }
      pattern = ((GroupPattern) pattern).getChild(childIndex);
      pathIndex++;
    }
  }

  private static final boolean isPatternOptional(NonEmptyPattern pattern) {
    if (pattern.getPatternType() != NonEmptyPattern.PATTERN_GROUP)
      return false;
    switch (((GroupPattern) pattern).getGroupType()) {
      case GroupPattern.GROUP_CHOICE_INCLUDING_EMPTY_PATTERN:
        return true;
      case GroupPattern.GROUP_CHOICE_EXCLUDING_EMPTY_PATTERN:
        for (int k = 0; k < ((GroupPattern) pattern).getChildCount(); k++)
          if (isPatternOptional(((GroupPattern) pattern).getChild(k)))
            return true;
        return false;
      default:
        for (int k = 0; k < ((GroupPattern) pattern).getChildCount(); k++)
          if (!isPatternOptional(((GroupPattern) pattern).getChild(k)))
            return false;
        return true;
    }
  }

  // Adds to a set all the element definitions that can be used
  //   for replacing an element that occupy the position given
  //   by a pattern path, relative to a given pattern.
  public static final void addElementDefinitionsForReplace(
    final Set elementDefinitionSet, final NonEmptyPattern rootPattern,
    final int[] patternPath) {
    NonEmptyPattern pattern = rootPattern;
    int pathIndex = 0;
    while (true) {
      int childIndex = patternPath[pathIndex];
      boolean isPatternPathEnd = childIndex < 0;
      boolean isLeafPattern =
        pattern.getPatternType() != NonEmptyPattern.PATTERN_GROUP;
      if (isLeafPattern != isPatternPathEnd)
        // The pattern path end must map to a leaf pattern
        throw new IllegalArgumentException();
      if (isPatternPathEnd)
        return;
      int groupType = ((GroupPattern)pattern).getGroupType();
      if (groupType == GroupPattern.GROUP_CHOICE_INCLUDING_EMPTY_PATTERN
        || groupType == GroupPattern.GROUP_CHOICE_EXCLUDING_EMPTY_PATTERN)
        for (int k = 0; k < ((GroupPattern)pattern).getChildCount(); k++)
          if (k != childIndex)
            addElementDefinitions(elementDefinitionSet,
              ((GroupPattern)pattern).getChild(k));
      pattern = ((GroupPattern)pattern).getChild(childIndex);
      pathIndex++;
    }
  }

  // Adds to a set all the element definitions that can be used
  //   for replacing an element that repeats the pattern of an element
  //   that occupy the position given by a pattern path,
  //   relative to a given pattern.
  public static final void addElementDefinitionsForReplaceRepeatedUse(
    final Set elementDefinitionSet, final NonEmptyPattern rootPattern,
    final int[] patternPath) {
    NonEmptyPattern pattern = rootPattern;
    boolean canBeRepeated = false;
    int pathIndex = 0;
    while (true) {
      int childIndex = patternPath[pathIndex];
      boolean isPatternPathEnd = childIndex < 0;
      boolean isLeafPattern =
        pattern.getPatternType() != NonEmptyPattern.PATTERN_GROUP;
      if (isLeafPattern != isPatternPathEnd)
        // The pattern path end must map to a leaf pattern
        throw new IllegalArgumentException();
      if (isPatternPathEnd)
        return;
      int groupType = ((GroupPattern)pattern).getGroupType();
      if (groupType == GroupPattern.GROUP_CHOICE_INCLUDING_EMPTY_PATTERN
        || groupType == GroupPattern.GROUP_CHOICE_EXCLUDING_EMPTY_PATTERN) {
        if (!canBeRepeated)
          canBeRepeated = pattern.canBeRepeated();
        if (canBeRepeated)
          for (int k = 0; k < ((GroupPattern)pattern).getChildCount(); k++)
            if (k != childIndex)
              addElementDefinitions(elementDefinitionSet,
              ((GroupPattern)pattern).getChild(k));
      }
      else
        canBeRepeated = false;
      pattern = ((GroupPattern)pattern).getChild(childIndex);
      pathIndex++;
    }
  }

  // Adds to a set all the allowed element definitions referenced by a pattern
  //   which can be inserted between two element definitions referenced by
  //   their pattern paths.
  // Doesn't take into account the element definitions
  //   which can replace the two base element definitions.
  public static final void addElementDefinitionsInsertableBetween(
    final Set elementDefinitionSet, final NonEmptyPattern rootPattern,
    final int[] patternPath1, final int[] patternPath2) {
    if (patternPath1 == null && patternPath2 == null) {
      addElementDefinitions(elementDefinitionSet, rootPattern);
      return;
    }
    if (patternPath1 == null) {
      addElementDefinitionsInsertableBefore(elementDefinitionSet, rootPattern,
        patternPath2, 0);
      return;
    }
    if (patternPath2 == null) {
      addElementDefinitionsInsertableAfter(elementDefinitionSet, rootPattern,
        patternPath1, 0);
      return;
    }
    NonEmptyPattern pattern = rootPattern;
    int pathIndex = 0;
    while (true) {
      int childIndex1 = patternPath1[pathIndex];
      int childIndex2 = patternPath2[pathIndex];
      boolean isLeafPattern =
        pattern.getPatternType() != NonEmptyPattern.PATTERN_GROUP;
      if (isLeafPattern || childIndex1 < 0 || childIndex2 < 0) {
        if (isLeafPattern && childIndex1 < 0 && childIndex2 < 0
            && childIndex1 == childIndex2)
          return;
        else // The pattern path end must map to a leaf pattern
          throw new IllegalArgumentException();
      }
      switch (((GroupPattern)pattern).getGroupType()) {
        case GroupPattern.GROUP_GROUP:
          if (childIndex1 > childIndex2)
            throw new IllegalArgumentException();
          if (childIndex1 < childIndex2) {
            addElementDefinitionsInsertableAfter(
                elementDefinitionSet,
                ((GroupPattern)pattern).getChild(childIndex1), patternPath1,
                pathIndex + 1);
            for (int k = childIndex1 + 1; k < childIndex2; k++)
              addElementDefinitions(
                  elementDefinitionSet, ((GroupPattern)pattern).getChild(k));
            addElementDefinitionsInsertableBefore(
                elementDefinitionSet,
                ((GroupPattern)pattern).getChild(childIndex2), patternPath2,
                pathIndex + 1);
            return;
          }
          break;
        case GroupPattern.GROUP_INTERLEAVE:
          for (int k = 0; k < ((GroupPattern)pattern).getChildCount(); k++) {
            if (k != childIndex1 && k!= childIndex2)
              addElementDefinitions(elementDefinitionSet,
              ((GroupPattern)pattern).getChild(k));
            else if (k != childIndex1)
              addElementDefinitionsInsertableBefore(elementDefinitionSet,
                  ((GroupPattern)pattern).getChild(childIndex2), patternPath2,
                  pathIndex + 1);
            else if (k != childIndex2)
              addElementDefinitionsInsertableAfter(elementDefinitionSet,
                  ((GroupPattern)pattern).getChild(childIndex1), patternPath1,
                  pathIndex + 1);
          }
        default:
          if (childIndex2 != childIndex1)
            return;
      }
      pattern = ((GroupPattern)pattern).getChild(childIndex1);
      pathIndex++;
    }
  }

  // Adds to a set all the allowed element definitions referenced by a pattern
  //   which can be inserted before an element definition referenced
  //   by its pattern path.
  // Doesn't take into account the element definitions
  //   which can replace the base element definition.
  private static final void addElementDefinitionsInsertableBefore(
    final Set elementDefinitionSet, final NonEmptyPattern rootPattern,
    final int[] patternPath, final int startIndex) {
    NonEmptyPattern pattern = rootPattern;
    int pathIndex = startIndex;
    while (true) {
      int childIndex = patternPath[pathIndex];
      boolean isPatternPathEnd = childIndex < 0;
      boolean isLeafPattern =
        pattern.getPatternType() != NonEmptyPattern.PATTERN_GROUP;
      if (isLeafPattern != isPatternPathEnd)
        // The pattern path end must map to a leaf pattern
        throw new IllegalArgumentException();
      if (isPatternPathEnd)
        return;
      int groupType = ((GroupPattern)pattern).getGroupType();
      if (groupType == GroupPattern.GROUP_GROUP
        || groupType == GroupPattern.GROUP_INTERLEAVE)
        for (int k = 0; k < childIndex; k++)
          addElementDefinitions(elementDefinitionSet,
          ((GroupPattern)pattern).getChild(k));
      if (groupType == GroupPattern.GROUP_INTERLEAVE)
        for (int k = childIndex + 1;
          k < ((GroupPattern)pattern).getChildCount(); k++)
          addElementDefinitions(elementDefinitionSet,
          ((GroupPattern)pattern).getChild(k));
      pattern = ((GroupPattern)pattern).getChild(childIndex);
      pathIndex++;
    }
  }

  // Adds to a set all the allowed element definitions referenced by a pattern
  //   which can be inserted after an element definition referenced
  //   by its pattern path.
  // Doesn't take into account the element definitions
  //   which can replace the base element definition.
  private static final void addElementDefinitionsInsertableAfter(
    final Set elementDefinitionSet, final NonEmptyPattern rootPattern,
    final int[] patternPath, final int startIndex) {
    NonEmptyPattern pattern = rootPattern;
    int pathIndex = startIndex;
    while (true) {
      int childIndex = patternPath[pathIndex];
      boolean isPatternPathEnd = childIndex < 0;
      boolean isLeafPattern =
        pattern.getPatternType() != NonEmptyPattern.PATTERN_GROUP;
      if (isLeafPattern != isPatternPathEnd)
        // The pattern path end must map to a leaf pattern
        throw new IllegalArgumentException();
      if (isPatternPathEnd)
        return;
      int groupType = ((GroupPattern)pattern).getGroupType();
      if (groupType == GroupPattern.GROUP_INTERLEAVE)
        for (int k = 0; k < childIndex; k++)
          addElementDefinitions(elementDefinitionSet,
          ((GroupPattern)pattern).getChild(k));
      if (groupType == GroupPattern.GROUP_GROUP
        || groupType == GroupPattern.GROUP_INTERLEAVE)
        for (int k = childIndex + 1;
          k < ((GroupPattern)pattern).getChildCount(); k++)
          addElementDefinitions(elementDefinitionSet,
          ((GroupPattern)pattern).getChild(k));
      pattern = ((GroupPattern)pattern).getChild(childIndex);
      pathIndex++;
    }
  }
}
