/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.util.dom;

import org.w3c.dom.Node;
import org.w3c.dom.Element;

/**
 * Utilities about empty texts and mixed content elements.
 */
public abstract class Cleaner {

  /**
   * Remove the empty test child nodes of parentNode, recursively.
   * An empty text may only have spaces and tabs.
   */
  public static final int removeEmptyTexts(Node parentNode) {
    int removeCount = 0;
    Node childNode = parentNode.getFirstChild();
    while (childNode != null) {
      Node nextChildNode = childNode.getNextSibling();
      switch (childNode.getNodeType()) {
        case Node.ELEMENT_NODE:
          removeCount += removeEmptyTexts(childNode);
          break;
        case Node.TEXT_NODE:
          if (isWhitespaceValue(childNode.getNodeValue())) {
            parentNode.removeChild(childNode);
            removeCount++;
          }
          break;
      }
      childNode = nextChildNode;
    }
    return removeCount;
  }

  /**
   * Return true if <code>value</value> contains only spaces and tabs,
   *   false elsewhere.
   */
  public static final boolean isWhitespaceValue(String value) {
    for (int k = 0; k < value.length(); k++) {
      char ch = value.charAt(k);
      if (ch != ' ' && ch != '\n' && ch != '\r' && ch != '\t')
        return false;
    }
    return true;
  }

  /**
   * Return the first element that has a mixed content.
   * Starts with the <code>rootElement</code>, then continue with its descendants
   *   until finding an element with a mixed content.
   * A mixed content has non empty text nodes *and* child elements.
   * Return null if no element has mixed content.
   */
  public static final Element mixedContentElement(Element rootElement) {
    boolean hasNonEmptyText = false;
    boolean hasChildElement = false;
    Node childNode = rootElement.getFirstChild();
    while (childNode != null) {
      switch (childNode.getNodeType()) {
        case Node.ELEMENT_NODE: {
          if (hasNonEmptyText)
            return rootElement;
          Element mixedContentElement = mixedContentElement((Element)childNode);
          if (mixedContentElement != null)
            return mixedContentElement;
          hasChildElement = true;
          break;
        }
        case Node.TEXT_NODE: {
          if (!isWhitespaceValue(childNode.getNodeValue())) {
            if (hasChildElement)
              return rootElement;
            hasNonEmptyText = true;
          }
          break;
        }
      }
      childNode = childNode.getNextSibling();
    }
    return null;
  }
}
