/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.util.dom;

import java.io.IOException;
import java.io.InputStream;

import org.w3c.dom.Element;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.InputSource;
import org.xml.sax.EntityResolver;
import org.xml.sax.ContentHandler;
import org.xml.sax.Locator;
import org.xml.sax.ErrorHandler;
import org.xml.sax.Attributes;
import org.apache.xerces.parsers.SAXParser;

import org.xmloperator.util.exception.ExceptionControler;
import org.xmloperator.util.dom.IntStack;
import org.xmloperator.util.dom.ElementBrowser;

public final class ElementLocator implements ContentHandler {

  public static Element locateElement(Element rootElement,
      final String systemId, final int lineNumber, final int columnNumber)
      throws IOException, SAXException {
    int[] elementIndexStack = locateElement(systemId, lineNumber, columnNumber);
    return fromElementIndexStackToElement(rootElement, elementIndexStack);
  }

  private static final class LocationFoundException extends SAXException {
    public LocationFoundException() {
      super("");
    }
  }
  private static final LocationFoundException LOCATION_FOUND_EXCEPTION =
      new LocationFoundException();

  // Return an element index stack
  public static int[] locateElement(
      final String systemId, final int lineNumber, final int columnNumber)
      throws IOException, SAXException {
    XMLReader parser = new SAXParser();
    IntStack indexStack = new IntStack();
    ElementLocator elementLocator =
        new ElementLocator(indexStack, lineNumber, columnNumber);
    parser.setContentHandler(elementLocator);
    parser.setErrorHandler(
      new ErrorHandler() {
        public void warning(SAXParseException exception) throws SAXException {
        }
        public void error(SAXParseException exception) throws SAXException {
          throw exception;
        }
        public void fatalError(SAXParseException exception)
            throws SAXException {
          throw exception;
        }
      });
    parser.setEntityResolver(new EntityResolver() {
      public InputSource resolveEntity (
          String publicId, String systemId)
          throws SAXException, IOException {
        return new InputSource(new InputStream() {
          public int read() throws IOException {
            return 0;
          }
        });
      }
    });
    try {
      parser.parse(systemId);
    }
    catch (LocationFoundException exception) {
      int[] indexes = new int[indexStack.size()];
      indexStack.copyInto(indexes);
      return indexes;
    }
    return  null;
  }

  // Return the element referenced by the elementIndexStack.
  public static final Element fromElementIndexStackToElement(
      final Element rootElement, final int[] elementIndexStack) {
    if (rootElement == null || elementIndexStack == null)
      return null;
    Element element = null;
    for (int depth = 0; depth < elementIndexStack.length; depth++) {
      if (element == null)
        element = rootElement;
      else {
        Element firstChildElement =
            ElementBrowser.getFirstChildElement(element);
        if (firstChildElement == null)
          return depth == elementIndexStack.length - 1 ? element : null;
        element = firstChildElement;
      }
      for (int k = 0; k < elementIndexStack[depth]; k++) {
        element = ElementBrowser.getNextSiblingElement(element);
        if (element == null)
          return null;
      }
    }
    return element;
  }

  private IntStack indexStack;
  private int lineNumber, columnNumber;
  private Locator locator;

  private ElementLocator(IntStack indexStack, int lineNumber, int columnNumber) {
    this.indexStack = indexStack;
    this.lineNumber = lineNumber;
    this.columnNumber = columnNumber;
    indexStack.push(0); // Top is expected, not yet realized.
  }

  ////////////////////////////////
  // ContentHandler implementation
  ////////////////////////////////

  public void setDocumentLocator (Locator locator) {
    this.locator = locator;
  }

  private boolean isLocationReached() {
    int reachedLineNumber = locator.getLineNumber();
    int reachedColumnNumber = locator.getColumnNumber();
    return reachedLineNumber > lineNumber
        || (reachedLineNumber == lineNumber
        && reachedColumnNumber > columnNumber);
  }

  public void startDocument () throws SAXException {
  }

  public void endDocument () throws SAXException {
  }

  public void startPrefixMapping (String prefix, String uri)
    throws SAXException {
  }

  public void endPrefixMapping (String prefix) throws SAXException {
  }

  public void startElement (String namespaceURI, String localName,
  String rawName, Attributes atts) throws SAXException {
    if (isLocationReached())
      throw LOCATION_FOUND_EXCEPTION;
    indexStack.push(0);
  }

  public void endElement (String namespaceURI, String localName,
    String rawName)	throws SAXException {
    indexStack.pop();
    if (isLocationReached())
      throw LOCATION_FOUND_EXCEPTION;
    indexStack.nextTop();
  }

  public void characters (char ch[], int start, int length)
    throws SAXException {
    if (isLocationReached()){
      indexStack.pop();
      throw LOCATION_FOUND_EXCEPTION;
    }
  }

  public void ignorableWhitespace (char ch[], int start, int length)
    throws SAXException {
  }

  public void processingInstruction (String target, String data)
    throws SAXException {
    if (isLocationReached()) {
      indexStack.pop();
      throw LOCATION_FOUND_EXCEPTION;
    }
  }

  public void skippedEntity (String name)	throws SAXException {
  }
}
