/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.util.dom;

import java.util.StringTokenizer;

import org.w3c.dom.Node;
import org.w3c.dom.Element;
import org.w3c.dom.Document;

public abstract class NodePathBuilder {

  ////////////////////////////////////
  // Build a path for accessing a node
  ////////////////////////////////////

  // Return the node index without regarding the node type.
  public static final int getNodeIndex(Node node) {
    if (node == null)
      throw new IllegalArgumentException();
    int index = 0;
    while (true) {
      node = node.getPreviousSibling();
      if (node == null)
        return index;
      index++;
    }
  }

  // Return the element index without regarding the element name.
  // Nodes from other types are discards.
  public static final int getElementIndex(Element element) {
    int index = 0;
    Node node = element;
    while (true) {
      node = node.getPreviousSibling();
      if (node == null)
        return index;
      if (node.getNodeType() == Node.ELEMENT_NODE)
        index++;
    }
  }

  // Return the index of the named element.
  // Elements with other names are discards.
  public static final int getNamedElementIndex(Element element) {
    String elementName = element.getNodeName();
    int index = 0;
    Node node = element;
    while (true) {
      node = node.getPreviousSibling();
      if (node == null)
        return index;
      if (node.getNodeType() == Node.ELEMENT_NODE
        && node.getNodeName().equals(elementName))
        index++;
    }
  }

  // Build a path using named element indexes, such as :
  // "elementName elementIndex ... elementName elementIndex".
  public static final String buildElementPathWithNames(Element element) {
    if (element == null)
      return "";
    StringBuffer sb = new StringBuffer();
    while (true) {
      if (sb.length() != 0)
        sb.insert(0, ' ');
      sb.insert(0, getNamedElementIndex(element));
      sb.insert(0, ' ');
      sb.insert(0, element.getNodeName());
      Node parentNode = element.getParentNode();
      if (parentNode == null)
        throw new RuntimeException("Node has no parent");
      if (parentNode.getNodeType() != Node.ELEMENT_NODE)
        break;
      element = (Element)parentNode;
    }
    return sb.toString();
  }

  ///////////////////////////////
  // Find a node from a node path
  ///////////////////////////////

  // Return the child node of parentNode, with index childNodeIndex. 
  public static final Node getChildNodeFromIndex(Node parentNode,
    int childNodeIndex) {
    if (childNodeIndex < 0)
      return null;
    Node node = parentNode.getFirstChild();
    while (node != null) {
      if (childNodeIndex == 0)
        return node;
      node = node.getNextSibling();
      childNodeIndex--;
    }
    return null;
  }

  // Return the child element of parentElement with index childElementIndex.
  public static final Element getChildElementFromElementIndex(
    Element parentElement, int childElementIndex) {
    return getChildElementFromElementIndex(parentElement, null,
      childElementIndex);
  }

  // Return the child element of parentElement with index childElementIndex.
  // If elementName is not null then only elements with name elementName
  // are taken in account.
  public static final Element getChildElementFromElementIndex(
    Element parentElement, String elementName, int childElementIndex) {
    Node childNode = parentElement.getFirstChild();
    while (childNode != null) {
      if (childNode.getNodeType() == Node.ELEMENT_NODE
        && (elementName == null
        || childNode.getNodeName().equals(elementName))) {
        if (childElementIndex == 0)
          return (Element)childNode;
        childElementIndex--;
      }
      childNode = childNode.getNextSibling();
    }
    return null;
  }

  // Return the element at the end of the path.
  // The path is such as : "elementIndex ... elementIndex".
  public static final Element getElementFromPath(Document document,
    String elementPath) {
    StringTokenizer elementPathTokenizer =
      new StringTokenizer(elementPath);
    Element element = document.getDocumentElement();
    boolean isDocumentElement = true;
    while (elementPathTokenizer.hasMoreElements()) {
      int elementIndex = -1;
      try {
        elementIndex =
          Integer.decode(elementPathTokenizer.nextToken()).intValue();
      }
      catch (NumberFormatException exception) {
      }
      if (elementIndex < 0)
        return null;
      if (isDocumentElement) {
        if (elementIndex != 0)
          return null;
        isDocumentElement = false;
      }
      else {
        element = getChildElementFromElementIndex(element, null, elementIndex);
        if (element == null)
          return null;
      }
    }
    return element;
  }

  // Return the node at the end of the path.
  // The path is such as : "elementIndex ... elementIndex nodeIndex".
  public static final Node getNodeFromPath(Document document,
    String elementPath, int nodeIndex) {
    if (!new StringTokenizer(elementPath).hasMoreElements())
      return nodeIndex >= 0 ?
        getChildNodeFromIndex(document, nodeIndex) : document;
    Element element = getElementFromPath(document, elementPath);
    if (element == null)
      return null;
    return nodeIndex >= 0 ? getChildNodeFromIndex(element, nodeIndex) : element;
  }

  // Return the element at the end of the path.
  // The path is such as :
  // "elementName elementIndex ... elementName elementIndex".
  public static final Element getElementFromElementPathWithNames(
    Document document, String elementPath) {
    StringTokenizer elementPathTokenizer = new StringTokenizer(elementPath);
    if (!elementPathTokenizer.hasMoreElements())
      return null;
    Element element = document.getDocumentElement();
    boolean isDocumentElement = true;
    while (elementPathTokenizer.hasMoreElements()) {
      String elementName = elementPathTokenizer.nextToken();
      int elementIndex;
      try {
        elementIndex =
          Integer.decode(elementPathTokenizer.nextToken()).intValue();
      }
      catch (NumberFormatException exception) {
        return null;
      }
      if (isDocumentElement) {
        if (!elementName.equals(element.getNodeName()) || elementIndex != 0)
          return null;
        isDocumentElement = false;
      }
      else {
        element = getChildElementFromElementIndex(element, elementName,
          elementIndex);
        if (element == null)
          return null;
      }
    }
    return element;
  }
}

